#!/usr/bin/python

################################################################
#
# Copyright (C) 2003,2004 Tez Kamihira
#
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
#

import random
import sys
import string
import re


#BS = '\\'
BS = '`'

def dsign(n):

    if n < 0: return -1
    if n > 0: return 1
    return 0

def mysign(n):

    if n < 0: return '-'
    if n > 0: return '+'
    return '0'

def my_min(n, m):

    if n == None:
        return m
    return min(n, m)

def my_max(n, m):

    if n == None:
        return m
    return max(n, m)

class canvas:

    ctrl_chars = {
        -1:{-1:BS, 0:'-',  1:'/' },
         0:{-1:'|',  0:None, 1:'|' },
         1:{-1:'/',  0:'-',  1:BS}
        }

    ctrl_chars2 = {

        # character matrix or... merely gag ?

        "+++0":'`',  "++++":BS,   "++0+":':',
        "+0+-":'\'', "+0+0":'-',  "+0++":'.',
        "+-+-":'/',  "+-+0":'.',  "+-0-":':',
        "0+-+":':',  "0+0+":'|',  "0+++":':',

        "0---":':',  "0-0-":'|',  "0-+-":':',
        "-+-0":'\'', "-+-+":'/',  "-+0+":':',
        "-0--":'`',  "-0-0":'-',  "-0-+":'.',
        "--0-":':',  "----":BS,   "---0":'.'
        }

    def __init__(self, mode = "text", script_name = ''):

        if mode != "text":
            print "mode %s : not supported yet." % mode
            sys.exit()

        self.obj = []
        self.min_x = None
        self.min_y = None
        self.max_x = None
        self.max_y = None

        self.mldic = [ {}, {} ]     # 0 ... yours->mine, 1 ... mine->yours.

        self.script_name = script_name
        self.url_loc = {}           #  { y:{[[x, orig, cnv], [x2, orig2, cnv2], ...]} ... }

    def h_line(self, x, y, end_ch = ''):

        [x1, x2] = x

        xx = x1

        if x1 < x2:
            while xx <= x2:
                self.setch(xx, y, '-')
                xx = xx + 1
        else:
            while xx >= x2:
                self.setch(xx, y, '-')
                xx = xx - 1

        if end_ch != '':
            self.setch(x2, y, end_ch)

    def v_line(self, x, y, end_ch = ''):

        [y1, y2] = y

        yy = y1

        if y1 < y2:
            while yy <= y2:
                self.setch(x, yy, '|')
                yy = yy + 1
        else:
            while yy >= y2:
                self.setch(x, yy, '|')
                yy = yy - 1

        if end_ch != '':
            self.setch(x, y2, end_ch)

    def draw_forked_arrow(self, p1, p2, p3):

        self.obj.append(['forked_arrow', p1, p2, p3])

    def draw_forked_arrow_emit(self, p1, p2, p3):

        '''
                p1
                |
             +--+--+
             |     |
            p2     p3

        '''

        delta = 2 
        [[x1, y1], [x2, y2], [x3, y3]] = [p1, p2, p3]

        y_tmp = min(y2, y3)
        y_middle = (y_tmp + y1) / 2

        self.v_line(x1 - delta, [y1, y_middle])
        self.v_line(x1 + delta, [y1, y_middle])

        self.h_line([x2, x1 - delta], y_middle)
        self.h_line([x3, x1 + delta], y_middle)

        self.v_line(x2, [y_middle+1, y2], 'V')
        self.v_line(x3, [y_middle+1, y3], 'V')

    def bresenham(self, x1, y1, x2, y2, end_ch):

        # DDA ... by Bresenham.

        [dx, dy] = [x2 - x1, y2 - y1]

        e = 0

        plots = []

        if [dx, dy] == [0, 0]:
            plots.append([x1, y1])

        elif abs(dx) > abs(dy):

            x2 = x2 + dsign(dx)
            dy = dy + dsign(dy)
            dx = dx + dsign(dx)

            y = y1
            for x in range(x1, x2, dsign(dx)):
                e = e + abs(dy)
                if e > abs(dx):
                    e = e - abs(dx)
                    y = y + dsign(dy)
                plots.append([x, y])
        else:

            y2 = y2 + dsign(dy)
            dy = dy + dsign(dy)
            dx = dx + dsign(dx)

            x = x1
            for y in range(y1, y2, dsign(dy)):
                e = e + abs(dx)
                if e > abs(dy):
                    e = e - abs(dy)
                    x = x + dsign(dx)
                plots.append([x, y])

        self.draw_context_dependent_line(plots, end_ch)

    def draw_context_dependent_line(self, plots, end_ch):

        if len(plots) == 1:

            [x, y] = plots[0]
            if end_ch != '':
                self.setch(x, y, end_ch)
            else:
                self.setch(x, y, '.')

        elif len(plots) == 2:

            [[x1, y1], [x2, y2]] = plots
            [dx, dy] = [x2 - x1, y2 - y1]

            ch = canvas.ctrl_chars[dx][dy]
            self.setch(x1, y1, ch)
            self.setch(x2, y2, ch)

        elif len(plots) >= 3:

            [[x1, y1], [x2, y2]] = [plots[0], plots[1]]
            [dx, dy] = [x2 - x1, y2 - y1]

            ch = canvas.ctrl_chars[dx][dy]
            self.setch(x1, y1, ch)

            [[x1, y1], [x2, y2]] = [plots[-1], plots[-2]]
            [dx, dy] = [x2 - x1, y2 - y1]

            if end_ch != '':
                ch = end_ch
            else:
                ch = canvas.ctrl_chars[dx][dy]
            self.setch(x1, y1, ch)

            for i in range(len(plots) - 1)[1:]:

                [[x0, y0], [x1, y1], [x2, y2]] = [plots[i-1], plots[i], plots[i+1]]
                [dx1, dy1, dx2, dy2] = [x1 - x0, y1 - y0, x2 - x1, y2 - y1]

                [dx1, dy1, dx2, dy2] = map(lambda x: mysign(x), [dx1, dy1, dx2, dy2])
                ch = canvas.ctrl_chars2[ dx1 + dy1 + dx2 + dy2 ]

                self.setch(x1, y1, ch)

    def box(self, x1, y1, x2, y2):

        self.obj.append(['box', x1, y1, x2, y2])

    def line(self, x1, y1, x2, y2, end_ch = ''):

        self.obj.append(['line', x1, y1, x2, y2, end_ch])

    def mline(self, x1, y1, x2, y2, type):

        self.obj.append(['mline', x1, y1, x2, y2])

        key = str(x2) + ':' + str(y2)

        v = [x1, y1, x2, y2, type]

        if type == '<':
            if not self.mldic[0].has_key(key):
                self.mldic[0][key] = [v]
            else:
                if self.mldic[0][key].count(v) == 0:
                    self.mldic[0][key].append(v)
        elif type == '>':
            if not self.mldic[1].has_key(key):
                self.mldic[1][key] = [v]
            else:
                if self.mldic[1][key].count(v) == 0:
                    self.mldic[1][key].append(v)
        else:
            raise ValueError

    def emit_mline_aux(self, phase):

        for i in [0, 1]:
            for k in self.mldic[i].keys():
                pt = self.mldic[i][k]
                if len(pt) == 1:
                    if phase == 0:
                        self.bresenham(pt[0][0], pt[0][1], pt[0][2], pt[0][3], pt[0][4])
                else:
                    if phase == 1:
                        www = 3
                        if i == 0:  # mine <- yours.
                            yy = map(lambda x: x[1], pt)
                            yy.sort()
                            self.v_line(pt[0][0] - www, [yy[0], yy[-1]])
                            map(lambda p: self.h_line([p[0], p[0] - www], p[1], '+'), pt)

                            end_y = pt[0][3]
                            if end_y < yy[0]:
                                best_y = yy[0]
                            elif end_y > yy[-1]:
                                best_y = yy[-1]
                            elif end_y + 1 >= yy[0] and end_y + 1 <= yy[-1]:
                                best_y = end_y + 1
                            else:
                                best_y = end_y

                            self.bresenham(pt[0][0] - www - 1, best_y, pt[0][2], end_y, '<')

                        else:       # mine -> yours.
                            yy = map(lambda x: x[1], pt)
                            yy.sort()
                            self.v_line(pt[0][0] + www, [yy[0], yy[-1]])
                            map(lambda p: self.h_line([p[0], p[0] + www], p[1], '+'), pt)

                            end_y = pt[0][3]
                            if end_y < yy[0]:
                                best_y = yy[0]
                            elif end_y > yy[-1]:
                                best_y = yy[-1]
                            elif end_y + 1 >= yy[0] and end_y + 1 <= yy[-1]:
                                best_y = end_y + 1
                            else:
                                best_y = end_y

                            self.bresenham(pt[0][0] + www + 1, best_y, pt[0][2], end_y, '>')

    def emit_mline(self):

        self.emit_mline_aux(0)
        self.emit_mline_aux(1)

    def mline2(self, slist, sp, ep, end_ch, width):

        self.obj.append(['mline2', slist, sp, ep, end_ch, width])

    def emit_mline2(self, slist, sp, ep, end_ch, width):

        if end_ch == '>':
            self.v_line(slist[0][0] + width, [slist[0][1], slist[-1][1]])
            map(lambda p: self.h_line([p[0], p[0] + width], p[1], '+'), slist)
            if sp == None:
                sp = [0, 0]
                sp[0] = slist[0][0] + width + 1
                sp[1] = slist[0][1]
            self.bresenham(sp[0], sp[1], ep[0], ep[1], end_ch)
        else:
            self.v_line(slist[0][0] - width, [slist[0][1], slist[-1][1]])
            map(lambda p: self.h_line([p[0], p[0] - width], p[1], '+'), slist)
            if sp == None:
                sp = [0, 0]
                sp[0] = slist[0][0] - width - 1
                sp[1] = slist[0][1]
            self.bresenham(sp[0], sp[1], ep[0], ep[1], end_ch)

    def text(self, x, y, s, b_frame = 1):

        if b_frame == 0:
            self.obj.append(['text', x, y, s])
        else:
            s = ' ' + s + ' '
            self.obj.append(['text', x + 1, y + 1, s])
            self.obj.append(['box', x, y, x + len(s) + 1, y + 2])

    def rich_center_text(self, x1, y1, x2, y2, s, s_orig = ''):

        if len(s) > x2 - x1 - 1:
            s = s[:x2 - x1 - 1]
        n_pad = (x2 - x1 - 1 - len(s)) / 2

        self.box(x1, y1, x2, y2)
        self.text(x1+1 + n_pad, y1+1, s, 0)

        if s_orig != '':
            if self.url_loc.has_key(str(y1+1)):
                self.url_loc[str(y1+1)].append([ x1 + 1 + n_pad, s, s_orig])
            else:
                self.url_loc[str(y1+1)] = [[x1 + 1 + n_pad, s, s_orig]]

    def extend_x(self, x):

        self.min_x = my_min(self.min_x, x)
        self.max_x = my_max(self.max_x, x)

    def extend_y(self, y):

        self.min_y = my_min(self.min_y, y)
        self.max_y = my_max(self.max_y, y)

    def extend_xy(self, x, y):

        self.extend_x(x)
        self.extend_y(y)

    def setch(self, x, y, ch):
        xx = x - self.min_x + self.mergin_x
        yy = y - self.min_y + self.mergin_y
        self.canvas[yy] = self.canvas[yy][:xx] + ch + self.canvas[yy][xx+1:]

    def emit(self):

        self.min_x, self.min_y = None, None
        self.max_x, self.max_y = None, None

#        self.mergin_x, self.mergin_y = 10, 10
        self.mergin_x, self.mergin_y = 1, 1
        self.extend_xy(0, 0)

        for i in range(len(self.obj)):

            obj_type = self.obj[i][0]

            if obj_type in ['box', 'line', 'mline']:
                self.extend_xy(self.obj[i][1], self.obj[i][2])
                self.extend_xy(self.obj[i][3], self.obj[i][4])
            elif obj_type == 'text':
                self.extend_xy(self.obj[i][1], self.obj[i][2])
                self.extend_x(self.obj[i][1] + len(self.obj[i][3]))
            elif obj_type == 'forked_arrow':
                self.extend_xy(self.obj[i][1][0], self.obj[i][1][1])
                self.extend_xy(self.obj[i][2][0], self.obj[i][2][1])
                self.extend_xy(self.obj[i][3][0], self.obj[i][3][1])
            elif obj_type == 'mline2':
                slist = self.obj[i][1]
                sp = self.obj[i][2]
                ep = self.obj[i][3]
                end_ch = self.obj[i][4]
                width = self.obj[i][5]

                map(lambda x: self.extend_xy(x[0], x[1]), slist)
                if sp != None:
                    self.extend_xy(sp[0], sp[1])
                self.extend_xy(ep[0], ep[1])

        a_line = ' ' * (self.max_x - self.min_x + self.mergin_x * 2)
        self.canvas = [a_line] * (self.max_y - self.min_y + self.mergin_y * 2)

        for i in range(len(self.obj)):

            obj_type = self.obj[i][0]

            if obj_type == 'box':
                x1 = self.obj[i][1]
                y1 = self.obj[i][2]
                x2 = self.obj[i][3]
                y2 = self.obj[i][4]

                for i in range(x2 - x1):
                    self.setch(x1 + i, y1, '-')
                    self.setch(x1 + i, y2, '-')
                for i in range(y2 - y1):
                    self.setch(x1, y1 + i, '|')
                    self.setch(x2, y1 + i, '|')
                self.setch(x1, y1, '+')
                self.setch(x1, y2, '+')
                self.setch(x2, y1, '+')
                self.setch(x2, y2, '+')

            elif obj_type == 'line':

                x1 = self.obj[i][1]
                y1 = self.obj[i][2]
                x2 = self.obj[i][3]
                y2 = self.obj[i][4]
                end_ch = self.obj[i][5]

                self.bresenham(x1, y1, x2, y2, end_ch)

            elif obj_type == 'forked_arrow':

                p1 = self.obj[i][1]
                p2 = self.obj[i][2]
                p3 = self.obj[i][3]
                self.draw_forked_arrow_emit(p1, p2, p3)

            elif obj_type == 'text':
                x = self.obj[i][1]
                y = self.obj[i][2]
                s = self.obj[i][3]

                for i in range(len(s)):
                    self.setch(x+i, y, s[i])

            elif obj_type == 'mline2':
                slist = self.obj[i][1]
                sp = self.obj[i][2]
                ep = self.obj[i][3]
                end_ch = self.obj[i][4]
                width = self.obj[i][5]
                self.emit_mline2(slist, sp, ep, end_ch, width)

        self.emit_mline()

        for i in range(len(self.canvas)):
            es = string.rstrip(self.canvas[i])

            ''' handle html mode '''
            if self.script_name != '':
                # print self.url_loc
                es = embed_hrefs(es, self.url_loc, self.min_x, self.min_y + i, self.script_name)

            print es


def embed_hrefs_aux(script_name, s1, s2):

    '''
        a--patch-1
           ==>    <a href="scriptname/A/C--B--V--patch-1">a--patch-1</a>

        a--patch-3..5
           ==>    <a href="scriptname/A/C--B--V--patch-3">a--patch-3</a>..
                  <a href="scriptname/A/C--B--V--patch-5">5</a>

           where A,C,B,V are archive,category,branch,version, respectively.

    '''

    m = re.match(r"^(.+)-(\d+)\.\.(\d+)$", s1)
    if m != None:
        s1_head = m.group(1)
        s1_digit1 = m.group(2)
        s1_digit2 = m.group(3)
        m = re.match(r"^(.+)-(\d+)\.\.(\d+)$", s2)
        if m == None:
            print s1, s2
            sys.exit(0)
        s2_head = m.group(1)
        s2_digit1 = m.group(2)
        s2_digit2 = m.group(3)

        s2_orig1 = s2_head + '-' + s2_digit1
        s2_orig2 = s2_head + '-' + s2_digit2
        return '<a href="%s/%s-%s" title="%s">%s-%s</a>..<a href="%s/%s-%s" title="%s">%s</a>' % \
                  (script_name, s2_head, s2_digit1, s2_orig1, s1_head, s1_digit1, script_name, s2_head, s2_digit2, s2_orig2, s1_digit2)
    else:
        return '<a href="%s/%s" title="%s">%s</a>' % (script_name, s2, s2, s1)

def embed_hrefs(line, dic, x_org, y_org, script_name):

    list_list = dic.get(str(y_org-1), [])        # XXX: !!!
    x_org = x_org - 1                            # XXX: !!!
    list_list.sort(lambda x, y: cmp(x[0], y[0]))

    x_bias = 0
    for i in range(len(list_list)):
        x = list_list[i][0] - x_org + x_bias
        s_cnv = list_list[i][1]
        s_org = list_list[i][2]

        ss = embed_hrefs_aux(script_name, s_cnv, s_org)

        line = line[:x] + ss + line[x + len(s_cnv):]

        x_bias = x_bias + len(ss) - len(s_cnv)

    return line

def _main():
             
    d = canvas("text")
#    d.box(0, 0, 10, 10)
#    d.box(15, 15, 30, 30)
#    d.box(16, 16, 17, 17)
#    d.text(1, 1, 'hello world')
#    d.line(15, 5, 1, 3)
#    d.text(10, 20, 'Tez')
#    d.rich_center_text(10, 10, 30, 15, 'XXX')
    d.line(1, 1, 80, 80)
    d.draw_forked_arrow([10, 10], [5, 14], [15, 14])
    d.mline2([[10, 10], [10, 15], [10, 20]], [14, 12], [20, 10], '>', 3)
    d.line(1,10, 20, 1)
    d.emit()


if __name__ == '__main__': _main()

# arch-tag: 67890c92-beba-4c1a-bb98-d9d5b3b49a4d
