#!/usr/bin/python

################################################################
#
# Copyright (C) 2003,2004 Tez Kamihira
#
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
#


import os
import tlatools
import vchart
import re
import string
import stat
try:
	import tempfile
	tempfile.mkstemp
except (ImportError, AttributeError):
	import vtempfile as tempfile

import compat

def path_name_comparator(f, g):

    return cmp(f, g)

def We_must_work_around_it_anyway(fname, if_map_d):

    path_elements = fname.split('/')
    if len(path_elements) <= 1:
        return

    dir_path = path_elements[0]
    b_found = 0
    for iv in if_map_d.keys():
        if if_map_d[iv] == dir_path:
            b_found = 1
            break

    if b_found == 0:
        if_map_d[ 'A_./' + dir_path ] = dir_path
        tlatools.tmsg('w intermediate dir %s not found. Fixed.' % dir_path)

    if len(path_elements) > 2:
        for i in range(len(path_elements) - 2):
            dir_path = dir_path + '/' + path_elements[ i + 1 ]
            b_found = 0
            for iv in if_map_d.keys():
                if if_map_d[iv] == dir_path:
                    b_found = 1
                    break
            if b_found == 0:
                if_map_d[ 'A_./' + dir_path ] = dir_path
                tlatools.tmsg('w intermediate dir %s not found. Fixed.' % dir_path)

class changeset:

    def __init__(self):

        self.patch_dir_root = None
        self.patch_dir = None

    def rd_idx_file(self, fname):

        fi_dic = {}      # file -> inventory
        if_dic = {}      # inventory -> file

        fname = tlatools.trim_dot_slash(fname)

        f = open(self.patch_dir + '/' + fname, 'r')
        line = f.readline()
        while line:
            m = re.match(r"^(\S+)\s+(\S+)$", line)
            if m:
                unescaped_f = tlatools.pika_unescape(m.group(1))
                ff = tlatools.trim_dot_slash(unescaped_f)
                ii = m.group(2)

                fi_dic[ff] = ii
                if_dic[ii] = ff

            line = f.readline()
        f.close()

        return [fi_dic, if_dic]

    def open(self, acbvr):

        self.patch_dir_root = tempfile.mkdtemp()
        self.patch_dir = self.patch_dir_root + '/patch'
        self.acbvr = acbvr
        tlatools.tsystem('tla get-patch %s %s' % (tlatools.del_null_branch(acbvr), self.patch_dir))

        [self.orig_dirs_index_fi_map,  self.orig_dirs_index_if_map]  = self.rd_idx_file('orig-dirs-index')
        [self.orig_files_index_fi_map, self.orig_files_index_if_map] = self.rd_idx_file('orig-files-index')
        [self.mod_dirs_index_fi_map,   self.mod_dirs_index_if_map]   = self.rd_idx_file('mod-dirs-index')
        [self.mod_files_index_fi_map,  self.mod_files_index_if_map]  = self.rd_idx_file('mod-files-index')

    def close(self):

        tlatools.rmdir_always(self.patch_dir_root)

    def apply_changeset(self, cf_f, cf_d, inv_dir, inv_wk_dir, seq_no):

        '''
           if the last transaction was incomplete, this function would be
           called with same seq_no again. To keep the cache consistent, seq_no check is essential.

              Do not make irresposible comments. Before this function is called, we always
              clean up working space and after that our state is always definitive.
              So this check does not have any sense. Do you understand ?

           I'm sorry.

        '''

        ''' load last dir list and expand to map '''
        lines = cf_d.check_out_to_list(seq_no - 1)
        if_map_d = {}
        for i in range(len(lines)):
            m = re.match(r"^(\S+)\s+(\S+).*$", lines[i])
            if m:
                if_map_d[ m.group(2) ] = tlatools.pika_unescape(m.group(1))

        ''' load last file list and expand to map '''
        lines = cf_f.check_out_to_list(seq_no - 1)
        if_map_f = {}
        for i in range(len(lines)):
            m = re.match(r"^(\S+)\s+(\S+).*$", lines[i])
            if m:
                if_map_f[ m.group(2) ] = tlatools.pika_unescape(m.group(1))

        ''' _remove_ files. '''
        for iv in self.orig_files_index_if_map.keys():
            if not self.mod_files_index_if_map.has_key(iv):

                # print "Remove : File : " + if_map_f[ iv ]

                try:
                    del if_map_f[ iv ]

                except Exception:
                    pass

        ''' _set aside_ renamed files. '''
        renamed_files_list = []
        for iv in self.orig_files_index_if_map.keys():
            if self.mod_files_index_if_map.has_key(iv):
                f_orig = self.orig_files_index_if_map[iv]
                f_mod = self.mod_files_index_if_map[iv]
                renamed_files_list.append([iv, f_orig, f_mod])

                # print "Rename : File : %s -> %s" % (f_orig, f_mod)

                try:
                    del if_map_f[ iv ]
                except Exception:
                    pass

        ''' _remove_ dirs. '''
        for iv in self.orig_dirs_index_if_map.keys():
            if not self.mod_dirs_index_if_map.has_key(iv):

                # print "Remove : dir : %s" % if_map_d[iv]

                try:
                    del if_map_d[ iv ]

                    ''' This adjustment is really needed ? '''
                    for ivf in if_map_f.keys():
                        if if_map_f[ivf][ : len(if_map_d[iv]) ] == if_map_d[iv]:
                            del if_map_f[ivf]

                    ''' This adjustment is really needed ? '''
                    for ivd in if_map_d.keys():
                        if if_map_d[ivd][ : len(if_map_d[iv]) ] == if_map_d[iv]:
                            del if_map_d[ivd]

                except Exception:
                    pass

        ''' _set aside_ dirs. '''
        renamed_dirs_list = []
        for iv in self.orig_dirs_index_if_map.keys():
            if self.mod_dirs_index_if_map.has_key(iv):
                d_orig = self.orig_dirs_index_if_map[iv]
                d_mod = self.mod_dirs_index_if_map[iv]

                # print "Rename : dir : %s -> %s" % (d_orig, d_mod)

                ''' This process is really needed ? '''
                files_under_dir = []
                for ivf in if_map_f.keys():

                    slash_pos = if_map_f[ivf].rfind('/')
                    if slash_pos >= 0 and if_map_f[ivf][ : slash_pos ] == if_map_d[iv]:

                        # print "files_under_dir : %s : %s(%s)" % ( if_map_d[iv], if_map_f[ivf], ivf )

                        files_under_dir.append([ivf, if_map_f[ivf]])
                        del if_map_f[ivf]

                '''
                    This process is really needed ?

                     ANSWER:  NO!
                '''
                dirs_under_dir = []
                # for ivd in if_map_d.keys():
                #     try:
                #         if if_map_d[ivd][ : len(if_map_d[iv]) ] == if_map_d[iv]:
                #             dirs_under_dir.append([ivd, if_map_d[ivd]])
                #             del if_map_d[ivd]
                #     except Exception:
                #         pass

                renamed_dirs_list.append([iv, d_orig, d_mod, files_under_dir, dirs_under_dir])

                try:
                    del if_map_d[ iv ]

                except Exception:
                    pass

        ''' restore rename/new directories. 
            Because in ViewARCH, this process is virtual,
            I can restore them by arbitrary order.
              ( not always shallow -> deep. )
        '''
        ''' rename directoreies '''
        for i in range(len(renamed_dirs_list)):
            [iv, d_orig, d_mod, files_under_dir, dirs_under_dir] = renamed_dirs_list[i]
            for j in range(len(files_under_dir)):
                ivf = files_under_dir[j][0]
                fnf = files_under_dir[j][1]
                fnf = d_mod + fnf[ len(d_orig) : ]
                if_map_f[ivf] = fnf
            for j in range(len(dirs_under_dir)):
                ivd = dirs_under_dir[j][0]
                fnd = dirs_under_dir[j][1]
                fnd = d_mod + fnd[ len(d_orig) : ]
                if_map_d[ivd] = fnd

            if_map_d[iv] = d_mod

        ''' new directoreies '''
        for iv in self.mod_dirs_index_if_map.keys():
            if not self.orig_dirs_index_if_map.has_key(iv):
                if_map_d[iv] = self.mod_dirs_index_if_map[iv]

        ''' restore rename/new files.
            I can restore them by arbitrary order.
              ( not always shallow -> deep. )
        '''
        ''' renamed files '''
        for i in range(len(renamed_files_list)):
            [iv, f_orig, f_mod] = renamed_files_list[i]
            if_map_f[iv] = f_mod

        ''' new files '''
        for iv in self.mod_files_index_if_map.keys():

            if not self.orig_files_index_if_map.has_key(iv):
                if_map_f[iv] = self.mod_files_index_if_map[iv]

                new_file_name = self.patch_dir + '/new-files-archive/'
                new_file_name = new_file_name + tlatools.trim_dot_slash(if_map_f[ iv ])

                f_inv_out = inv_wk_dir + '/' + tlatools.iv2iv64(iv) + '.gz'
                f_inv_in = inv_dir + '/' + tlatools.iv2iv64(iv) + '.gz'
                cf_inv = vchart.vchart()
                cf_inv.load(f_inv_in)

                cf_inv.check_in_from_file(new_file_name, tlatools.trim_dot_slash(if_map_f[ iv ]), seq_no, None)

                cf_inv.save(f_inv_out)

                #####################################################################################
                ##  At least in some cases, tla does not register intermediate dir in mod-dirs-index,
                ##  even if much more deeper regular files are appended.
                ##  This would be a mkpatch's bug. We must work around it anyway.
                #####################################################################################

                We_must_work_around_it_anyway(tlatools.trim_dot_slash(if_map_f[ iv ]), if_map_d)

                #####################################################################################
                ##  End of work around.
                #####################################################################################

        ''' make fi_map from if_map for sort processing '''
        fi_map_d = {}
        for iv in if_map_d.keys():
            fi_map_d[ if_map_d[iv] ] = iv

        ''' sort result list and make output lines '''
        files = fi_map_d.keys()
        files.sort(path_name_comparator)
        out_lines = []
        for i in range(len(files)):
            out_lines.append('%s\t%s\n' % ( tlatools.pika_escape(files[i]), fi_map_d[ files[i] ] ))

        ''' write back to vchart '''
        cf_d.check_in_from_vanilla_list(out_lines, 'dir_list', seq_no)

        ''' flush vchart '''
        ''' Do not flush it. It's upper function's business. (XXX: more clearly.) '''
        # cf_d.save()

        ''' make fi_map from if_map for sort processing '''
        fi_map_f = {}
        for iv in if_map_f.keys():
            fi_map_f[ if_map_f[iv] ] = iv

        ''' sort result list and make output lines '''
        files = fi_map_f.keys()
        files.sort(path_name_comparator)
        out_lines = []
        for i in range(len(files)):
            out_lines.append('%s\t%s\n' % ( tlatools.pika_escape(files[i]), fi_map_f[ files[i] ] ))

        ''' write back to vchart '''
        cf_f.check_in_from_vanilla_list(out_lines, 'file_list', seq_no)

        ''' flush vchart '''
        ''' Do not flush it. It's upper function's business. (XXX: more clearly.) '''
        # cf_f.save()

        # XXX: FILE-PERMISSION-HISTORY

        ''' process normal patch files '''
        patch_prefix_len = len(self.patch_dir + '/patches/')
        for root, dirs, files in compat.walk(self.patch_dir + '/patches'):
            for i in range(len(files)):
                if files[i][-6:] == '.patch':
                    real_f = files[i][:-6]

                    try:
                        abs_f_name = root + '/' + real_f
                        fn = abs_f_name[ patch_prefix_len : ]
                        # print fn
                        iv = self.mod_files_index_fi_map[ fn ]
                        # print iv
                        iv64 = tlatools.iv2iv64(iv)

                        f_inv_out = inv_wk_dir + '/' + iv64 + '.gz'
                        f_inv_in = inv_dir + '/' + iv64 + '.gz'
                        cf_inv = vchart.vchart()
                        cf_inv.load(f_inv_in)
    
                        # print "check in %s (%s)" % (fn, iv)

                        cf_inv.patch_file(abs_f_name + '.patch', seq_no)
                        cf_inv.set_last_commit_revision(seq_no)
                        cf_inv.file_name_history.putval(seq_no, fn)
                        cf_inv.file_type_history.putval(seq_no, 'T')

                        cf_inv.save(f_inv_out)

                    except KeyError, key:
                        tlatools.tmsg('w %s : bad key. still continue.' % str(key))

    def stat(self):

        [t_minus, t_plus, b_minus, b_plus] = [0, 0, 0, 0]

        for root, dirs, files in compat.walk(self.patch_dir + '/patches'):
            for i in range(len(files)):
                abs_f_name = root + '/' + files[i]
                if files[i][-6:] == '.patch':

                    fp = open(abs_f_name)
                    mm = pp = 0
                    ln = fp.readline()
                    while ln:
                        if len(ln) > 0:
                            if ln[0] == '-':
                                mm = mm + 1
                            elif ln[0] == '+':
                                pp = pp + 1
                        ln = fp.readline()
                    fp.close()

                    if mm > 0:
                        t_minus = t_minus + (mm - 1)
                    if pp > 0:
                        t_plus = t_plus + (pp - 1)

                elif files[i][-9:] == '.original':
                    if tlatools.f_type(abs_f_name) == 'B':
                        b_minus = b_minus + os.stat(abs_f_name)[stat.ST_SIZE]

                elif files[i][-9:] == '.modified':
                    if tlatools.f_type(abs_f_name) == 'B':
                        b_plus = b_plus + os.stat(abs_f_name)[stat.ST_SIZE]

        for root, dirs, files in compat.walk(self.patch_dir + '/new-files-archive'):
            for i in range(len(files)):
                abs_f_name = root + '/' + files[i]
                if tlatools.f_type(abs_f_name) == 'T':

                    fp = open(abs_f_name)
                    lns = fp.readlines()
                    fp.close()
                    t_plus = t_plus + len(lns)

                else:
                    b_plus = b_plus + os.stat(abs_f_name)[stat.ST_SIZE]

        for root, dirs, files in compat.walk(self.patch_dir + '/removed-files-archive'):
            for i in range(len(files)):
                abs_f_name = root + '/' + files[i]
                if tlatools.f_type(abs_f_name) == 'T':

                    fp = open(abs_f_name)
                    lns = fp.readlines()
                    fp.close()
                    t_minus = t_minus + len(lns)

                else:
                    b_minus = b_minus + os.stat(abs_f_name)[stat.ST_SIZE]

        return [t_minus, t_plus, b_minus, b_plus]

    def get_patch_log(self):

        plog = self.patch_dir + '/new-files-archive/' + tlatools.get_patch_path(self.acbvr)
        if os.access(plog, os.F_OK):
            sz = os.stat(plog)[stat.ST_SIZE]
            f = open(plog)
            b = f.read(sz)
            f.close()
            return b
        else:
            print "%s : not found" % plog
            print "patch_dir : %s" % self.patch_dir
            print "acbvr : %s" % self.acbvr
            return ""

    def dump(self):
    
        print '------------------- orig_dirs_index ------------------------'
        for i in self.orig_dirs_index_fi_map.keys():
            print "%s -- %s" % (i, self.orig_dirs_index_fi_map[i])
        print '------------------- orig_files_index ------------------------'
        for i in self.orig_files_index_fi_map.keys():
            print "%s -- %s" % (i, self.orig_files_index_fi_map[i])
        print '------------------- mod_dirs_index ------------------------'
        for i in self.mod_dirs_index_fi_map.keys():
            print "%s -- %s" % (i, self.mod_dirs_index_fi_map[i])
        print '------------------- mod_files_index ------------------------'
        for i in self.mod_files_index_fi_map.keys():
            print "%s -- %s" % (i, self.mod_files_index_fi_map[i])

def _main():

    pass

if __name__ == '__main__': _main()

# arch-tag: b1dadb2a-efa9-4a40-878e-1cabe9f74597
