#!/usr/bin/python

################################################################
#
# Copyright (C) 2003,2004 Tez Kamihira
#
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
#

#########################################################################
#
# gauge_rev -- gauge revision loader.
#

import tlatools
import vchart
import re
import os
import stat

gauge_rev_dir = '/gauge-rev'
dtmp_file = '/dtmp'
ftmp_file = '/ftmp'

get_srcfind_RE = re.compile(r"^S\s+(\S+)\s+(\S+).*$")

def get_srcfind(wd, outf, b_dir):

    if b_dir == 1:
        ch = '-d'
    else:
        ch = '-f'

    g = open(outf, 'w')
    f = tlatools.topen('tla srcfind %s --all --ids . | sort' % ch, wd)
    line = f.readline()
    while line:

        m = re.match(get_srcfind_RE, line[:-1])
        if m != None:
            fname = tlatools.pika_unescape(m.group(1))
            iname = m.group(2)
            fname = tlatools.trim_dot_slash(fname)
            g.write('%s\t%s\n' % (tlatools.pika_escape(fname), iname))
        line = f.readline()

    f.close()
    g.close()

class gauge_rev:

    def __init__(self):

        self.bdir = None

    def open(self, acbvr, wdir):

        self.bdir = wdir + gauge_rev_dir
        self.acbvr = acbvr
        self.dtmp = wdir + dtmp_file
        self.ftmp = wdir + ftmp_file
        tlatools.mkdir_always(wdir)

        tlatools.tmsg('* %s' % self.acbvr)
        tlatools.tsystem('tla get -s --no-pristine %s %s' % (self.acbvr, self.bdir))

    def close(self):

        tlatools.rmdir_always(self.bdir)
        os.remove(self.dtmp)
        os.remove(self.ftmp)

    def get_gauge_revision(self, cf_d, cf_f, outdir):

        get_srcfind(self.bdir, self.dtmp, 1)
        cf_d.check_in_from_file(self.dtmp, "dir_list", 0, 'T')

        get_srcfind(self.bdir, self.ftmp, 0)
        cf_f.check_in_from_file(self.ftmp, "file_list", 0, 'T')

        inv_n = cf_f.check_out_to_list(0)
        self.get_initial_inventories(self.bdir, outdir, inv_n)

    def get_initial_inventories(self, tree_dir, inv_wk_dir, inv_n):

        [n_map, n_map_r] = tlatools.list2maps(inv_n)

        for iv in n_map_r.keys():

            f_inv_out = inv_wk_dir + '/' + tlatools.iv2iv64(iv) + '.gz'
            f_file = tree_dir + '/' + tlatools.trim_dot_slash(n_map_r[iv])

            cf_inv = vchart.vchart()

            cf_inv.check_in_from_file(f_file, n_map_r[iv], 0, None)
            cf_inv.save(f_inv_out)

    def get_patch_log(self):

        plog = self.bdir + '/' + tlatools.get_patch_path(self.acbvr)
        if os.access(plog, os.F_OK):
            sz = os.stat(plog)[stat.ST_SIZE]
            f = open(plog)
            b = f.read(sz)
            f.close()
            return b
        else:
            return ""


def _main():

    pass

if __name__ == '__main__': _main()

# arch-tag: 372ebe55-f55b-4424-8ae7-e1d46e64ffb2
