#!/usr/bin/python

################################################################
#
# Copyright (C) 2003,2004 Tez Kamihira
#
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
#

LIB_DIR = None

import sys

if LIB_DIR:
  sys.path.insert(0, LIB_DIR)


import os
import tlatools
from tlatools import cfg
import re
import getopt
import string
import stat
import tempfile
import vcache
import summary
import gauge_rev
import revmap

opt_verbose = 0
opt_force_cache = 0

def gen_vcache(super_root, archive, cbvr):

    [a, c, b, v, r] = tlatools.split_acbvr(cbvr)
    cbv = c + '--' + b + '--' + v
    if r != None:
        stop_rev = archive + '/' + cbv + '--' + r
    else:
        stop_rev = ''

    VCACHE = vcache.vcache(super_root, archive, cbv, None)

    if opt_force_cache == 0 and VCACHE.lock() == 0:
        tlatools.tmsg('! %s/%s is locked. skip.' % (archive, cbv))
        return

    ''' read recent summary '''
    sm = summary.summary(super_root + "/,,summary.txt")

    ''' check last incomplete transaction '''
    if VCACHE.need_roll_back():
        VCACHE.roll_back()
    elif VCACHE.need_roll_forward():
        VCACHE.roll_forward()

    VCACHE.load_index()
    vrev_name = VCACHE.get_next_vrev_name()
    vrev_id = VCACHE.get_next_vrev_id()
    while vrev_name != None:

        VCACHE.create_working_space()

        if vrev_id == 0:
            VCACHE.get_gauge_revision()
        else:
            VCACHE.apply_changeset()

        VCACHE.flush_revmap()

        VCACHE.advance_state()

        VCACHE.move_working_files_to_definitive_area()
        VCACHE.remove_working_stuff()

        VCACHE.advance_state()

        if vrev_name == stop_rev:
            break

        vrev_name = VCACHE.get_next_vrev_name()
        vrev_id = VCACHE.get_next_vrev_id()

    ''' update summary '''
    tm = tlatools.get_last_modified_time(archive, cbv)
    sm.append(tm, archive, cbv)
    sm.flush()

    VCACHE.unlock()

def check_vcache(super_root, archive, cbv):

    cbvr = string.split(cbv, '--')
    if len(cbvr) == 4:
        stop_rev = archive + '/' + cbv
        cbv = '--'.join(cbvr[:3])
    else:
        stop_rev = None

    vcache_dir = super_root + '/' + archive + '/' + cbv
    wd_tmp_dir = tempfile.mkdtemp('.ViewARCH', ',,', '/var/tmp')
    tlatools.rmdir_always(wd_tmp_dir)
    tlatools.mkdir_always(wd_tmp_dir)

    revmap = vcache.load_revmap(vcache_dir)
    if revmap.getsize() == 0:
        tlatools.tmsg("Warning: no revisions in %s/%s. exit." % (archive, cbv))
        return # no revisions.

    cf_d = vcache.load_index(vcache_dir, 'D')
    cf_f = vcache.load_index(vcache_dir, 'F')

    if vcache.cache_incomplete(vcache_dir):   # incomplete state
        tlatools.tmsg("Warning: incomplete state. exit.")
        return

    [major, minor] = vcache.get_state(vcache_dir)

    for i in range(revmap.getsize())[:major]:

        if i == 0:
            tlatools.tmsg('c %s' % revmap.getitem(i))
            tlatools.tsystem('tla get -s --no-pristine %s %s/tree' % (revmap.getitem(i), wd_tmp_dir))
        else:
            tlatools.tmsg('c %s' % revmap.getitem(i))
            tlatools.tsystem('tla replay -d %s/tree %s > /dev/null' % (wd_tmp_dir, revmap.getitem(i)))

        if opt_verbose:
            print 'check index.d[%d]' % i, 
        gauge_rev.get_srcfind(wd_tmp_dir + '/tree', wd_tmp_dir + '/d.tmp', 1)

        cf_d.check_out_to_file(i, '%s/d2.tmp' % wd_tmp_dir)
        tlatools.tsystem('diff -L tom/dir -L tez/dir -u d.tmp d2.tmp', wd_tmp_dir)

        if opt_verbose:
            print ' index.f[%d]' % i
        gauge_rev.get_srcfind(wd_tmp_dir + '/tree', wd_tmp_dir + '/f.tmp', 0)

        cf_f.check_out_to_file(i, '%s/f2.tmp' % wd_tmp_dir)
        tlatools.tsystem('diff -L tom/files -L tez/files -u f.tmp f2.tmp', wd_tmp_dir)

        inv_n = cf_f.check_out_to_list(i)
        check_vcache_charts(wd_tmp_dir, vcache_dir, inv_n, i)

        if stop_rev != None and revmap.getitem(i) == stop_rev:
            break

    tlatools.rmdir_always(wd_tmp_dir)

def check_vcache_charts(wd_tmp_dir, vcache_dir, inv_n, seq):

    [n_map, n_map_r] = tlatools.list2maps(inv_n)

    for iv in n_map_r.keys():

        #
        # Set up viewarch side's file into tez_file.
        #
        cf_inv = vcache.load_inv_by_iname(vcache_dir, iv)
        tez_file = '%s/fff.tmp' % wd_tmp_dir
        cf_inv.check_out_to_file(seq, tez_file)


        #
        # Set up tla side's file into tom_file.
        #
        tom_file = wd_tmp_dir + '/tree/' + tlatools.trim_dot_slash(n_map_r[iv])

        #
        # if tom_file is not symlink nor access, print error and goto next file.
        #
        if not stat.S_ISLNK(os.lstat(tom_file)[stat.ST_MODE]) and \
           not os.access(tom_file, os.F_OK):

            tlatools.tmsg('e %s : not found.' % tom_file)
            continue

        if stat.S_ISLNK(os.lstat(tom_file)[stat.ST_MODE]):
            #
            # if tom_file is symlink, put the distination to tom_file2.
            #
            tom_file2 = '%s/lll.tmp' % wd_tmp_dir
            linkf = os.readlink(tom_file)
            lk = open(tom_file2, 'w')
            lk.write('%s\n' % linkf)
            lk.close()
            tlatools.tmsg('i %s : symlink -> %s' % (tom_file, linkf))
            #tlatools.tsystem('cat %s' % tez_file)
        else:
            # tom_file2 = tom_file.
            tom_file2 = tom_file

        f = tlatools.topen('diff -L tom%s -L tez%s -u %s %s' % (tom_file, tom_file, tom_file2, tez_file))
        line = f.readline()
        if opt_verbose == 1 and len(line) == 0:
            print "No difference: '%s'" % tom_file

        # XXX: FILE-PERMISSION-HISTORY

        b_first = 1
        while line:
           if b_first == 1:
               if line[ : len('Binary')] == 'Binary':
                   if opt_verbose:
                       print "skip binary file '%s'" % tom_file
                   break
               else:
                   print "inventory: %s" % iv
                   # print "encoded iv: %s" % f_inv_in
                   b_first = 0

           print line,
           line = f.readline()

        f.close()

def usage():

    cache_dir = cfg.general.tla_cachedir

    sys.stderr.write(\
'''
Usage:
       gencache.py
       gencache.py -h

            show this help

       gencache.py [-x|-X] [-c] [-v] [-Z] [-d DIR] [-A archive] [C--B--V]
       gencache.py [-x|-X] [-c] [-v] [-Z] [-d DIR] [A/C--B--V]
       gencache.py -k [-d DIR]
       gencache.py -K [all] [-d DIR] [-A archive] [C--B--V]
       gencache.py -K [all] [-d DIR] [A/C--B--V]

            Either -x or -c must be specified, or -k or -K.
              If -x, cache will be actually generated (or updated).
              If -c, already existed cache will be integrity checked.
              If -k, print lock status of each version and exit.
              If -K, break locks.

            You can specify -x and -c simultaniously. In that case,
              generating cache first, then it will be checked.

            Using -d, you can specify distination directory,
              otherwise, '%s' will be used.

            if -v, outputs will be verbose.

            if -X instead of -x, ignore LOCK file.

            if -A, archive is specified explicitly.

            If A/C--B--V is specified, only that part will be checked
              ( or generated/updated ), otherwise, all of the archives
               in '.arch-params/=locations' are used, where A, C, B, V
              are Archive, Category, Branch, Version, respectively.

            if -Z, enter deep survey mode.

            Moreover, you can specify A/C--B--V--R instead of A/C--B--V where R is
              a revision. In that case, execution will be stopped at the revison.

Example:

       python gencache.py -x -d ./tmpcache a@b.com--2003/aa--bb--1.0
       python gencache.py -c -d ./tmpcache a@b.com--2003/aa--bb--1.0
       python gencache.py -x -d ./tmpcache -A a@b.com--2003 aa--bb--1.0
       python gencache.py -x -d ./tmpcache aa--bb--1.0  ( my-default-archive is used )

       python gencache.py -x
       python gencache.py -c


''' % cache_dir)


def check_deep_survey_cache(s, a, c, b, v):

    acbv = "%s/%s--%s--%s" % (a, c, b, v)
    acbv_tmp = "%s/,,%s--%s--%s" % (a, c, b, v)

    vcache_dir = "%s/%s" % (s, acbv)
    vcache_tmp_dir = "%s/%s" % (s, acbv_tmp)

    if os.access(vcache_dir, os.F_OK):
        print """
Directory %s already exists.
Do you want to erase it and regenerate fresh deep survey cache (Y)?
or simply updated old cache (N)?
(Y/N) """,
        y_or_n = sys.stdin.readline()
        if y_or_n[0] in ['Y', 'y']:
            print "move %s to temporal dir %s and removeing..." % (vcache_dir, vcache_tmp_dir),
            os.rename(vcache_dir, vcache_tmp_dir)
            tlatools.rmdir_always(vcache_tmp_dir)
            print "done."

def _main():

    if len(sys.argv) == 1:
        usage()
        sys.exit(2)

    try:
        opts, args = getopt.getopt(sys.argv[1:], "hxcvd:A:XkK:Z", \
                 [
                   "help",
                   "execute",
                   "integrity-check",
                   "verbose",
                   "directory",
                   "archive",
                   "force",
                   "show-lock-status",
                   "kill-locks",
                   "deep-survey"
                 ])
    except getopt.GetoptError:
        usage()
        sys.exit(2)

    ###########################################################################
    # parse options
    ###########################################################################
    global opt_verbose
    global opt_force_cache

    opt_verbose = 0

    opt_exec = 0
    opt_check = 0
    opt_force_cache = 0
    opt_show_lock_status = 0
    opt_kill_locks = 0

    tlatools.b_cmdline = 1

    super_root = cfg.general.tla_cachedir

    archive_name = None

    revmap.opt_deep_survey = 0
    for o, a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit()
        if o in ("-x", "--execute"):
            opt_exec = 1
        if o in ("-c", "--integrity-check"):
            opt_check = 1
        if o in ("-d", "--directory"):
            super_root = a
        if o in ("-v", "--verbose"):
            opt_verbose = 1
        if o in ("-A", "--archive"):
            archive_name = a
        if o in ("-X", "--force"):
            opt_force_cache = 1
            opt_exec = 1
        if o in ("-k", "--show-lock-status"):
            opt_show_lock_status = 1
        if o in ("-Z", "--deep-survey"):
            revmap.opt_deep_survey = 1
        if o in ("-K", "--kill-locks"):
            opt_kill_locks = 1
            if a != "all":
                print 'sorry, now kill lock only supports "all" option.'
                sys.exit()

    if opt_show_lock_status == 1:
        vcache.show_lock_status(super_root)
        sys.exit()

    if opt_kill_locks == 1:
        vcache.kill_locks(super_root)
        sys.exit()

    if [opt_exec, opt_check] == [0, 0]:
        sys.stderr.write("Neither '-x' nor '-c' specified.\n")
        sys.exit(2)

    if revmap.opt_deep_survey == 1 and len(args) == 0:
        sys.stderr.write("In deep survey mode, you must specify one version strictly (A/C--B--V).\n")
        sys.exit(2)

    tlatools.tmsg('--- START ---')
    if len(args) == 0:

        if archive_name != None:
            archives = [archive_name]
        else:
            archives = tlatools.get_archives(1)

        for i in range(len(archives)):
            versions = tlatools.get_versions_from_archive(archives[i])
            for j in range(len(versions)):

                if opt_exec == 1:
                    gen_vcache(super_root, archives[i], versions[j])

                if opt_check == 1:
                    check_vcache(super_root, archives[i], versions[j])

    elif len(args) == 1:

        [a, c, b, v, r] = tlatools.split_acbvr(args[0])

        """if -A is set, we take it forcely."""
        if archive_name != None:
            a = archive_name

        """check for deep survey mode."""
        if revmap.opt_deep_survey == 1:
            check_deep_survey_cache(super_root, a, c, b, v)

        if r != None:
            cbv_r = '%s--%s--%s--%s' % (c, b, v, r)
        else:
            cbv_r = '%s--%s--%s' % (c, b, v)

        if opt_exec == 1:
            gen_vcache(super_root, a, cbv_r)

        if opt_check == 1:
            check_vcache(super_root, a, cbv_r)

    else:
        usage()

    tlatools.tmsg('--- END ---')

if __name__ == '__main__': _main()

# arch-tag: f38b0ddc-2d27-4929-ba55-c2fedff6b551
