#!/usr/bin/python

################################################################
#
# Copyright (C) 2004 Tez Kamihira
#
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
#


import sys
import os
import string
import re
import stat

########################################################################
#
# gzt.py -- GPL replacement of Greg Stein's "ezt.py".
#
# Why do I have to re-invent such a neat stuff again... just stupid...
#
########################################################################

#
# Life is short. I use simple recursive descent way. -Tez
#
# ( "Aho, Sethi, Ullman: Compilers Principles, Techniques, and Tools" 
#     P.44, 82, 181-182, 736, 740 )
#

def read_file(fn):

    sz = os.stat(fn)[stat.ST_SIZE]
    f = open(fn)
    s = f.read(sz)
    f.close()

    return s

def file_expander(s, dir_path, env):

    m = re.match(r"^(.*?)\[include\s+([^]]+)\](.*)$", s, re.DOTALL)
    if m:
        s0 = m.group(1)
        fname = m.group(2)
        s2 = m.group(3)

        if fname[0] == '"':
            f_abs = dir_path + "/" + fname[1:-1]
        else:
            f_abs = dir_path + "/" + env.getvar_valueu(fname)
        s1 = read_file(f_abs)

        return s0 + file_expander(s1 + s2, dir_path, env)
    else:
        return s


def lex(s):

    """return next token. 
    qual_name is considered as a part of plain string, when parsing phase.

    according to -g, the syntax is as follows:

       [QUAL_NAME]
       [include "filename"]  or [include QUAL_NAME]
       [for QUAL_NAME] ... [end]
       [if-any QUAL_NAME] ... [else] ... [end]
       [if-index INDEX_FROM_FOR odd] ... [else] ... [end]
       [if-index INDEX_FROM_FOR even] ... [else] ... [end]
       [if-index INDEX_FROM_FOR first] ... [else] ... [end]
       [if-index INDEX_FROM_FOR last] ... [else] ... [end]
       [if-index INDEX_FROM_FOR NUMBER] ... [else] ... [end]
       [is QUAL_NAME STRING] ... [else] ... [end]
       [is QUAL_NAME QUAL_NAME] ... [else] ... [end]
    """

    if len(s) == 0:
        return None

    m = re.match(r"^(.*?)(\[(if |for |if-index |if-any |is |else|end|#)[^]]*\])(.*)", s, re.DOTALL)
    if m:
        s0 = m.group(1)
        if len(s0) > 0:
            return ['string', s0, s[len(s0):]]
        s1 = m.group(2)            # [if ...]
        s2 = m.group(3).strip()    # if
        s3 = m.group(4)            # rest.

        # split s1.
        s1 = s1[1:-1].split()

        # skip comment.
        if s1[0] == '#':
            return lex(s3)
        else:
            return [s2, s1, s3]

    return ['string', s, ""]

def parse(s, stop_sym=[]):

    result = []
    token = lex(s)
    while token != None:
        [type, value, rest] = token
        if type == 'string':
            result.append(value)
        elif type == 'for':
            [body, rest, type] = parse(rest, ['end'])
            result.append([value, body])
        elif type in ['if-any', 'if-index', 'is']:
            [body1, rest, type] = parse(rest, ['else', 'end'])
            if type == 'else':
                [body2, rest, type] = parse(rest, ['end'])
                result.append([value, body1, body2])
            elif type == 'end':
                result.append([value, body1, []])
        elif type in stop_sym:
            break
        token = lex(rest)

    return [result, rest, type]

def dump(tree, level):

    for i in range(len(tree)):
        if type(tree[i]) is str:
            print (' ' * level) + tree[i]
        else:
            subtree = tree[i]
            if len(subtree) == 2:
                print (' ' * level),
                print subtree[0]
                dump(subtree[1], level + 1)
            elif len(subtree) == 3:
                print (' ' * level),
                print subtree[0]
                dump(subtree[1], level + 1)
                dump(subtree[2], level + 1)
            else:
                raise Exception

def replace(s, env):

    """ replace qual_names in emit phase."""
    out = ""
    m = re.match(r"^(.*?)\[([\.a-zA-Z0-9_]+)\](.*)", s, re.DOTALL)
    if m:
        s0 = m.group(1)
        if len(s0) > 0:
            out = s0
        out = out + env.getvar_value(m.group(2))
        return out + replace(m.group(3), env)
    else:
        return s

def emit(tree, channel, env):

    """dump() and emit() are quite similar, but
    how to write a higher order function in python...?"""

    for i in range(len(tree)):
        if type(tree[i]) is str:
            channel.write(replace(tree[i], env))
        elif len(tree[i]) >= 2:
            subtree = tree[i]
            directive = subtree[0]
            op = directive[0]
            if op == 'for':
                v = env.getvar_value(directive[1])
                env.pushvar(directive[1], v)
                for j in range(len(v)):
                    emit(subtree[1], channel, env)
                    env.incvar(directive[1])
                env.popvar(directive[1])
            elif op == 'if-any':
                qual_name = env.getvar_value(directive[1])
                if type(qual_name) == int:
                    qual_name = str(qual_name)
                if qual_name == None or len(qual_name) == 0:
                    emit(subtree[2], channel, env)   # else part.
                else:
                    emit(subtree[1], channel, env)   # if part.
            elif op == 'if-index':
                index_from_for = env.getvar_index(directive[1])
                odd_even_first_last_num = directive[2]
                if odd_even_first_last_num == 'odd' and \
                   index_from_for % 2 == 1:
                    emit(subtree[1], channel, env)
                elif odd_even_first_last_num == 'even' and \
                     index_from_for % 2 == 0:
                    emit(subtree[1], channel, env)
                elif odd_even_first_last_num == 'first' and \
                     index_from_for == 0:
                    emit(subtree[1], channel, env)
                elif odd_even_first_last_num == 'last' and \
                     index_from_for == env.getvar_size(directive[1]) - 1:
                    emit(subtree[1], channel, env)
                elif index_from_for == odd_even_first_last_num:
                    emit(subtree[1], channel, env)
                else:
                    emit(subtree[2], channel, env)
            elif op == 'is':
                qual_name = directive[1]
                qual_str = directive[2]
                if len(qual_str) >= 2 and qual_str[0] == '"':
                    v2 = qual_str[1:-1]
                else:
                    v2 = env.getvar_value(qual_str)
                v1 = env.getvar_value(qual_name)
                if v1 == v2:
                    emit(subtree[1], channel, env)
                else:
                    emit(subtree[2], channel, env)
        else:
            raise Exception

class env:  # variable environment.

    def __init__(self, data):

        self.gvars = data
        self.lvars = {}

    def pushvar(self, varname, var):

        if self.lvars.has_key(varname):
            self.lvars[varname].append([var, 0])
        else:
            self.lvars[varname] = [[var, 0]]

    def popvar(self, varname):

        if self.lvars.has_key(varname):
            self.lvars[varname].pop()
            if len(self.lvars[varname]) == 0:
                del self.lvars[varname]
        else:
            raise Exception

    def getvar_value(self, varname):

        """
        if self.lvars.has_key(varname):  # local var.
            [v, i] = self.lvars[varname][-1]
            return v[i]
        """

        """support inexact longest matching."""
        match_len = 0
        the_key = None
        keys = self.lvars.keys()
        for i in range(len(keys)):
            if len(varname) >= len(keys[i]) and \
               varname[: len(keys[i])] == keys[i] and \
               match_len < len(keys[i]):
                match_len = len(keys[i])
                the_key = keys[i]

        if match_len > 0:   # local var. found.
            [v, i] = self.lvars[the_key][-1]
            vi = v[i]
            if match_len == len(varname):
                return vi
            else:
                remain = varname[ match_len + 1 : ]
                varlist = remain.split('.')
                if len(varlist) > 0:
                    for i in range(len(varlist)):
                        vi = vi.__dict__[varlist[i]]
                return vi
                
        else: # global var.
  
            varlist = varname.split('.')
            v = self.gvars[varlist[0]]
            if len(varlist) > 1:
                for i in range(len(varlist))[1:]:
                    v = v.__dict__[varlist[i]]
            return v

    def getvar_index(self, varname):

        return self.lvars[varname][-1][1]

    def incvar(self, varname):

        idx = self.lvars[varname][-1][1]
        self.lvars[varname][-1][1] = idx + 1

class Template:

    def __init__(self, template_path):

        if os.access(template_path, os.F_OK):
            self.template_path = template_path
            self.lines = read_file(template_path)
            ridx = self.template_path.rfind('/')
            if ridx >= 0:
                self.dir_path = self.template_path[:ridx]
            else:
                self.dir_path = "."
        else:  # str. debug case.
            if template_path[0] != '|':
                raise Exception
            self.template_path = None
            self.lines = template_path[1:] # actually string itself.
            self.dir_path = '.'

    def generate(self, channel, data):

        e = env(data)               # construct variable context.

        ''' expand INCLUDE directive, at first. '''
        s = file_expander(self.lines, self.dir_path, e)

        ''' expand the rest of directives. '''
        [tree, rest, type] = parse(s)     # parse tree.
        emit(tree, channel, e)      # fire.
        channel.write('\n')         # XXX:

def _main():

    flist = ["cs.tpl", "footer.tpl", "header.tpl", "list.tpl",
             "patch_set.tpl", "plain.tpl", "revisions.tpl"]

    data = {}
    data['taco'] = ['1', '2', '3']
    data['imo'] = ['aaa', 'bbb', 'ccc']
    data['a'] = [1, 2, 3]

    template = Template(',,tp.txt')
    #template = Template("../templates/default/" + flist[0])
    template.generate(sys.stdout, data)

    #print test1, "----->"
    #[tree, rest, type] = parse(test1)

    #e = env(data)
    #emit(tree, e)
    #print '\n'
    #dump(tree, 0)

    #template = ezt.Template(os.path.join(tlatools.pathinfo_install_dir(), tn))
    #template = Template("../templates/default/" + flist[0])
    #template.generate(sys.stdout, data)

    
if __name__ == '__main__': _main()

# arch-tag: 3adcd6b2-a4f9-4554-a564-d818053d3599

