#!/usr/bin/python

################################################################
#
# Copyright (C) 2004 Tez Kamihira
#
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
#


#########################################################################
#
# pikaesc.py -- pika esc.
#

import string
import re

dic1 = {
    '\t': "\\(tab)",
    '\n': "\\(nl)",
    '\r': "\\(cr)",
    '\f': "\\(np)",
    ' ': "\\(sp)",
    '\\': "\\\\",
    '"': "\\\""
}

def pika_esc(s):

    out = ""
    for i in range(len(s)):

        if dic1.has_key(s[i]):
            out = out + dic1[s[i]]
        else:
            ch = ord(s[i])

            if (ch >= 0 and ch <= 31) or \
               (ch == 32)             or \
               (ch >= 128):

                # control or space or non-ascii.
                out = out + ("\\(U+%02X)" % ch)
            else:
                out = out + s[i]

    return out

def stox_(x):

    od = ord(x)
    if od >= ord('0') and od <= ord('9'):
         return od - ord('0')
    if od >= ord('a') and od <= ord('f'):   # allowable ?
         return od - ord('a') + 10
    if od >= ord('A') and od <= ord('F'):
         return od - ord('A') + 10

    return 0

def stox(xx):

    val = stox_(xx[0]) * 16 + stox_(xx[1])

    return val

dic2 = {
    "(tab)":'\t',
    "(nl)":'\n',
    "(cr)":'\r',
    "(np)":'\f',
    "(sp)":' ',
    "\\":'\\',
    "\"": '"'
}

def pika_unesc(s):

    out = ""
    while (len(s) > 0):
        if s[0] != '\\':
            out = out + s[0]
            s = s[1:]
        else:
            s = s[1:]

            m = re.match(r"^(\(tab\)|\(nl\)|\(cr\)|\(np\)|\(sp\)|\\|\")", s)
            if m:
                ss = m.group(1)
                if dic2.has_key(ss):
                    out = out + dic2[ss]
                    s = s[ len(ss) : ]
                else:
                    out = out + '\\'
            else:
                m = re.match(r"^\(U\+([A-F0-9][A-F0-9])\)", s)
                if m:
                    out = out + chr(stox(m.group(1)))
                    s = s[ 6: ]
                else:
                    out = out + '\\'

    return out

def test1():

    for i in range(256):
        s = chr(i)
        if s != pika_unesc(pika_esc(s)):
             print "error : %d\n" % i


def _main():

    test1()

if __name__ == '__main__': _main()

# arch-tag: ab439b99-2baa-46a5-9967-2de75422a84a
