#!/usr/bin/python


################################################################
#
# Copyright (C) 2004 Tez Kamihira
#
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
#

LIB_DIR = None

import sys

if LIB_DIR:
  sys.path.insert(0, LIB_DIR)

import os
import urllib
import re
import viewarch

import tlatools
from tlatools import cfg
from tlatools import html_quote
import accept

import gettext

_ = gettext.gettext


class request:

    def __init__(self):

        """
        for i in os.environ.keys():
            tlatools.msg('%s : %s' % (i, os.environ[i]))
        """

        ############################################################
        # Derived from LiteCGI.pm in ArchZoom.

	# work around Apache sending QUERY_STRING as parameter to cgi
        if os.getenv('SCRIPT_NAME') != None:
	    sys.argv = []

        query_string = ''
        if len(sys.argv) >= 3:
            query_string = sys.argv[2]
        else:
            query_string = os.getenv('QUERY_STRING','')

        path_info = ''
        if len(sys.argv) >= 2:
            path_info = sys.argv[1]
        else:
            path_info = os.getenv('PATH_INFO', '')
        if len(path_info) >= 1 and path_info[0] != '/':
            path_info = '/' + path_info

        if os.getenv('SCRIPT_NAME') != None:
            script_name = os.environ['SCRIPT_NAME']
        elif len(sys.argv) >= 1:
            script_name = sys.argv[0]
        else:
            script_name = ''
        if len(script_name) >= 1 and script_name[0] != '/':
            script_name = '/' + script_name

        #
        ############################################################

        #self.path_info = urllib.unquote(os.getenv('PATH_INFO', ''))
        #self.script_name = os.getenv('SCRIPT_NAME')

        self.path_info = urllib.unquote(path_info)
        self.script_name = script_name
        self.server_name = os.getenv('SERVER_NAME', '???')

        hal = os.getenv('HTTP_ACCEPT_LANGUAGE','')
        self.lang_selector = accept.language(hal)
        self.language = self.lang_selector.select_from(cfg.general.languages)
        self.b_dsm = 0

        #qs = os.getenv('QUERY_STRING')
        qs = query_string
        if qs:
            qs_split = qs.split('&')
        else:
            qs_split = []

        self.params = {}
        for i in range(len(qs_split)):
            p = qs_split[i].split('=')
            if len(p) == 1:
                self.params[p[0]] = ''
            elif len(p) == 2:
                self.params[p[0]] = urllib.unquote(p[1])

        if self.params.has_key('d'):
            tlatools.b_debug = 1

        if self.path_info in [None, '', '/']:
            self.type = viewarch.RQ_LIST_ARCHIVE            # initial exec.
            return

        if self.path_info[-1] == '/':
            b_trailing_slash = 1
            self.path_info = self.path_info[:-1]
        else:
            b_trailing_slash = 0

        names = self.path_info.split('/')

        if len(names) == 2:      # archive selected.
            self.archive = names[1]
            if self.params.has_key('abr'):
                self.type = viewarch.RQ_LIST_ABROWSE
            else:
                self.type = viewarch.RQ_LIST_CATEGORY
            return

        if len(names) >= 3:    # archive + c/b/v/r + dir selected.
            self.archive = names[1]
            cbvrd = names[2].split('--')
            if len(cbvrd) == 1: # category selected.
                self.type = viewarch.RQ_LIST_BRANCH
                self.category = cbvrd[0]
                return

            if len(cbvrd) == 2: # branch selected.
                self.type = viewarch.RQ_LIST_VERSION
                self.category = cbvrd[0]
                self.branch = cbvrd[1]
                return

            if len(cbvrd) == 3: # version selected.
                if self.params.has_key('graph'):
                    self.type = viewarch.RQ_DRAW_GRAPH
                else:
                    self.type = viewarch.RQ_LIST_PATCH_LOG
                self.category = cbvrd[0]
                self.branch = cbvrd[1]
                self.version = cbvrd[2]
                self.revision = ''
                return

            if len(cbvrd) == 4: # revision selected.
                self.category = cbvrd[0]
                self.branch = cbvrd[1]
                self.version = cbvrd[2]
                self.revision = cbvrd[3]

                if self.revision == "dsm":

                    """ using viewarch.cgi/a/c--b--v--dsm,
                    we mean listing patch logs for all vrevs.
                    """
                    # XXX: check whether dsc exists or not.

                    self.type = viewarch.RQ_LIST_PATCH_LOG
                    self.category = cbvrd[0]
                    self.branch = cbvrd[1]
                    self.version = cbvrd[2]
                    self.revision = ''
                    self.b_dsm = 1

                    return

                """check other deep survey mode."""
                if self.revision[:len("dsm-")] == "dsm-":
                    self.b_dsm = 1
                    self.vrev_no = int(self.revision[len("dsm-"):])

                    # check_valid_vrev(...)

                else:

                    if tlatools.check_valid_revision(self.archive,
                                                     self.category,
                                                     self.branch,
                                                     self.version,
                                                     self.revision) == 0:
                        raise Exception, _("Revision : %s/%s--%s--%s--%s not found") % (
                                            self.archive,
                                            self.category,
                                            self.branch,
                                            self.version,
                                            self.revision)

                if b_trailing_slash == 1:  # browse source tree's root directory.
                    if self.params.has_key('iv'):
                        self.type = viewarch.RQ_LIST_INV_HISTORY
                    else:
                        self.type = viewarch.RQ_ITEM_BROWSE
                    self.dirpart = './'
                elif len(names) >= 4:      # browse subdir or normal file.
                    if self.params.has_key('iv'):
                        self.type = viewarch.RQ_LIST_INV_HISTORY
                    else:
                        self.type = viewarch.RQ_ITEM_BROWSE
                    self.dirpart = './' + '/'.join(names[3:])
                else:
                    cmd_array = {
                        'vp': viewarch.RQ_GET_ALL_PATCHES,
                        'cs_mod':  viewarch.RQ_CAT_CSET_FILE,
                        'cs_new':  viewarch.RQ_CAT_CSET_FILE,
                        'cs_rm':  viewarch.RQ_CAT_CSET_FILE,
                        'graph':  viewarch.RQ_DRAW_GRAPH
                    }
                    for k in cmd_array.keys():
                        if self.params.has_key(k):
                            self.type = cmd_array[k]
                            return
                    self.type = viewarch.RQ_CAT_PATCH_LOG

    def add_navigator_element(self, anchor, s, classstr = ''):
	self.navigator_nest_lvl += 1
	if classstr != '':
	    classstr=' class="' + classstr + '"'
	return '<ul><li%s><a href="%s">%s</a>' % (classstr, anchor, s)

    def finish_navigator_list(self, s):
	while self.navigator_nest_lvl > 0:
	   s +=  '</li></ul>'
	   self.navigator_nest_lvl -= 1
	return s

    def make_navigator(self):

        s = self.__dict__.get('script_name', '')
        a = self.__dict__.get('archive', '')
        c = self.__dict__.get('category', '')
        b = self.__dict__.get('branch', '')
        v = self.__dict__.get('version', '')
        r = self.__dict__.get('revision', '')

	if s == '':
            return ''

	self.navigator_nest_lvl = 0

        ret = self.add_navigator_element(s, _('view archives'))
        if a == '':
            return self.finish_navigator_list(ret)

        ret += self.add_navigator_element('/'.join([s, a]), a)
        if c == '':
            return self.finish_navigator_list(ret)

        ret += self.add_navigator_element('/'.join([s, a, c]), c)
        if b == '':
            return self.finish_navigator_list(ret)

        ret += self.add_navigator_element('/'.join([s, a, c]) + '--' + b, b, "hyphen")
        if v == '':
            return self.finish_navigator_list(ret)

        ret += self.add_navigator_element('/'.join([s, a, c]) + '--'.join(['', b, v]), v, "hyphen")
	if r == '':
            return self.finish_navigator_list(ret)

        ret += self.add_navigator_element('/'.join([s, a, c]) + '--'.join(['', b, v, r]), r, "hyphen")
        return self.finish_navigator_list(ret)


# arch-tag: 42caa9ce-f73f-4f71-aa9b-c56d7f131750
