#!/usr/bin/python

################################################################
#
# Copyright (C) 2003,2004 Tez Kamihira
#
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
#


#########################################################################
#
# INSTALL-TIME CONFIGURATION
#
# These values will be set during the installation process. During
# development, they will remain None.
#

CONF_ARCH_PATHNAME = None

import sys
import os
import string
import re
import pprint
import stat
import time
import base64
import md5
import pikaesc
try:
    import datetime
except (ImportError, AttributeError):
    pass

import config

try:
    import tempfile
    tempfile.mkstemp
except (ImportError, AttributeError):
    import vtempfile as tempfile

import compat

b_debug = 0
b_cmdline = 0

def msg(s):

    pp = pprint.PrettyPrinter()
    ss = pp.pformat(s)

    f = open("/tmp/viewarch.log", "a")
    f.write(ss + '\n')
    f.close()


if CONF_ARCH_PATHNAME:
    conf_arch_pathname = CONF_ARCH_PATHNAME
else:
    abs_path = os.path.abspath('.')
    par_path = os.path.dirname(abs_path)
#    par_path = os.path.dirname(par_path)

    conf_arch_pathname = par_path + '/viewarch.conf'


cfg = config.Config()
cfg.load_config(conf_arch_pathname)


def pathinfo_install_dir():

    return os.path.dirname(conf_arch_pathname)

def tmnow():

    try:
        tm = str(datetime.datetime(2000, 1, 1).now())
        tm = tm[:-3] + '.' + tm[-3:]
    except Exception:
        tm = time.strftime('%Y/%m/%d %H:%M:%S')

    return tm

def tmsg(s, b_log = 0):

    tm = tmnow()
    if b_log != 0:
        msg('[%s] %s' % (tm, s))
    else:
        print '[%s] %s' % (tm, s)

def f_type(fn):

    if stat.S_ISLNK(os.lstat(fn)[stat.ST_MODE]):
        return 'S'    # symbolic link.

    elif stat.S_ISDIR(os.lstat(fn)[stat.ST_MODE]):
        return 'D'    # directory

    else:
        ''' for simplicity, I check only first 8K bytes of input file.
            if there are any 0x00, consider it 'binary'.

            if file size == 0, consider it as 'text'.
        '''
        f = open(fn, 'r')
        bf = f.read(1024 * 8)
        f.close()

        if len(bf) > 0 and bf.find('\0') >= 0:
            return 'B'    # binary

        else:
            return 'T'    # text

def last_char(fn):

    f = open(fn, 'r')
    f.seek(-1, 2)
    last_ch = f.read(1)
    f.close()

    return ord(last_ch[0])

def remove_last_char(fn):

    sz = os.stat(fn)[stat.ST_SIZE]

    if sz > 0:
        f = open(fn, 'a+')
        f.truncate(sz - 1)
        f.close()


def split_creator_email(s):

    s = s.strip()
    m = re.match(r"^([^<]+)(<[^>]+>)$", s)
    if m:
        return [m.group(1).strip(), m.group(2).strip()]
    else:
        return [s, '']

def round_sec(s):

    m = re.match(r"^(.*)(\d\d:\d\d):\d\d(.*)$", s)
    if m:
        return m.group(1) + m.group(2) + m.group(3)
    else:
        return s

b_dir_hash = 1    # 1 if md5() hashing. otherwise 0.

def iv2iv64(iv):

    iv64 = base64.encodestring(iv)
    iv64 = string.replace(iv64, '\n', '')

    if b_dir_hash != 0:

        m = md5.new()
        m.update(iv64)
        s_hash = m.hexdigest()[:2]

        iv64 = '%s/%s' % (s_hash, iv64)

    return iv64

'''
   iv64_2iv() and iv2iv64() aren't symmetric. hmmm...
'''

def iv64_2iv(iv64):

    return base64.decodestring(iv64)

RE_two_columns = re.compile(r"^(\S+)\s+(\S+).*$")

def find_assoc(lines, p_key):

    p_key = pika_escape(p_key)

    for i in range(len(lines)):
        m = RE_two_columns.match(lines[i])
        if m != None:
            key = m.group(1)
            value = m.group(2)
            if key == p_key:  # exits.
                return value

    return None

RE_non_space = re.compile(r"^(\S+)")

def get_files_by_directory_name(lines, dirname, b_full, f_type):

    ''' dirname is real filename. not pika escaped. '''
    ''' lines include "<pika escaped filename> <its inventory>" list. '''

    dirname = pika_escape(dirname)

    if dirname[-1] != '/':      # foo/bar -> foo/bar/
        dirname = dirname + '/'
    if dirname[:2] == './':     # ./foo/bar -> foo/bar
        dirname = dirname[2:]

    files = []
    for i in range(len(lines)):
        m = RE_non_space.match(lines[i])
        if m:
            f_name = m.group(1)

            """when full expansion mode, skip arch control dir.
               because it includes too many files. """
            if b_full and f_name[:len('{arch}')] == '{arch}':
                continue

            if f_name.find(dirname) == 0:
                if b_full:
                    files.append([f_type, pika_unescape(f_name)])
                else:
                    f_name2 = f_name[len(dirname):]
                    if f_name2.find('/') == -1:
                        files.append([f_type, pika_unescape(f_name2)])

    files.sort(lambda x, y: cmp(x[1], y[1]))
    return files

def trim_dot_slash(s):

    if len(s) >= 2 and s[:2] == './':
        return s[2:]
    else:
        return s

RE_digit1 = re.compile(r"^(\d+)$")
RE_digit2 = re.compile(r"^(\d+)\.\.(\d+)$")

def get_dd(dd_str):

    m = RE_digit1.match(dd_str)
    if m:
        d = int(m.group(1))
        return [d, d]

    m = RE_digit2.match(dd_str)
    if m:
        return [ int(m.group(1)), int(m.group(2)) ]

    return [None, None]

RE_two_columns0 = re.compile(r"^(\S+)\s+(\S+)$")

def list2maps(list, b_pika_unescape = 1):

    [map, map_r] = [{}, {}]

    for i in range(len(list)):
        m = RE_two_columns0.match(list[i])
        c1 = m.group(1)
        if b_pika_unescape == 1:
            c1 = pika_unescape(c1)
        c2 = m.group(2)
        [map[c1], map_r[c2]] = [c2, c1]

    return [map, map_r]

def add_null_branch(cbv_r):

    ret_list = string.split(cbv_r, '--')
    if len(ret_list) == 2 and ret_list[1] != 'NULL':
        ret_list.insert(1, 'NULL')

    return '--'.join(ret_list)

def del_null_branch(cbv_r):

    ret_list = string.split(cbv_r, '--')
    if len(ret_list) >= 2 and ret_list[1] == 'NULL':
        del ret_list[1]

    return '--'.join(ret_list)

RE_split_acbvr = re.compile(r"^([^/]+)/([^/]+)$")

def split_acbvr(acbvr):

    m = RE_split_acbvr.match(acbvr)
    if m:
        a = m.group(1)
        cbvr = m.group(2)
    else:
        a = my_default_archive()
        cbvr = acbvr

    ret_list = string.split(cbvr, '--')
    ret_list.insert(0, a)

    if len(ret_list) > 5:
        ret_list = ret_list[:5]

    while len(ret_list) < 5:
        ret_list.append(None)

    return ret_list

def get_patch_path(acbvr):
    
    '''
    As first approximation, assume below schema.

    {arch}/C/C--B/C--B--V/A/patch-log/R
    '''

    [a, c, b, v, r] = split_acbvr(acbvr)

    if r == None:
        r = "base-0"

    return "{arch}/%s/%s--%s/%s--%s--%s/%s/patch-log/%s" % \
                  ( c, c,  b, c,  b,  v, a,           r)

def rmdir_always(d):

    # Delete everything reachable from the directory named in 'top'.
    # CAUTION:  This is dangerous!  For example, if top == '/', it
    # could delete all your disk files.

    try:
        for root, dirs, files in compat.walk(d, topdown=False):
            for name in files:
                os.remove(os.path.join(root, name))
            for name in dirs:
                try:
                    os.rmdir(os.path.join(root, name))
                except OSError:
                    os.remove(os.path.join(root, name))

        os.rmdir(d)

    except Exception:
        pass

def mkdir_always(dir):

    '''
          /a/b/c
          ./a/b/c
          a/b/c

          ../../a/b   <- not supported.
    '''

    if len(dir) > 0 and dir[-1] == '/':
        dir = dir[:-1]

    if len(dir) > 0 and dir[0] == '/':
        dir = dir[1:]
        b_abs = 1
    else:
        b_abs = 0

    if len(dir) >= 2 and dir[:2] == './':
        dir = dir[2:]

    paths = string.split(dir, '/')

    if len(paths) > 0 and b_abs == 1:
        paths[0] = '/' + paths[0]

    m_dir = ''
    for i in range(len(paths)):
        if i == 0:
            m_dir = paths[0]
        else:
            m_dir = m_dir + '/' + paths[i]

        if not os.access(m_dir, os.F_OK):
            try:
                os.mkdir(m_dir)
            except OSError, val:
                '''
                   17 means 'File exists' error.
                   This error can occurr at very low probability,
                   when more than two instance 'mkdir' simultainiously.
                '''
                if val[0] != 17:
                    raise Exception


def move_always(s, d):   # s and d must be on the same device.

    if not os.access(s, os.F_OK):
        return   # fail.

    if d[-1] == '/':
        return None

    idx = d.rfind('/')
    if idx >= 0:
        mkdir_always(d[:idx])

    #print "s(%s), d(%s)" % (s, d)
    os.rename(s, d)

def move_all_files_recursive(s_dir, d_dir):

    for root, dirs, files in compat.walk(s_dir):
        for name in files:
            s = os.path.join(root, name)
            d = os.path.join( d_dir + root[ len(s_dir) : ], name)
            #print "s(%s), d(%s)" % (s, d)
            move_always(s, d)

def mkstemp_always(suffix, prefix, dir):

    mkdir_always(dir)
    return tempfile.mkstemp(suffix, prefix, dir)

def make_tempfile_in_the_same_dir(f_name):

    rindex = f_name.rfind('/')
    if rindex == -1:
        dir = './'
    else:
        dir = f_name[ : rindex ]
    mkdir_always(dir)
    [f, ftmp] = tempfile.mkstemp('', ',,', dir)
    os.close(f)

    return ftmp

def topen(cmd, wdir=''):

    global b_debug
    global b_cmdline

    if wdir != '':
        save_dir = os.getcwd()
        os.chdir(wdir)
        
    # hack: reap all children
    try:
        while os.waitpid(-1, os.WNOHANG) >= 0:
            pass
    except OSError:
        pass

    g_homedir = cfg.general.tla_homedir
    if g_homedir != '':
        home_env = 'HOME=%s ' % g_homedir
    else:
        home_env = ''

    tla_abspath = cfg.general.tla_abspath
    if tla_abspath != '':
        if cmd[:4] == 'tla ':
            cmd = tla_abspath + cmd[3:]

    if b_cmdline == 0 and b_debug == 0:
        # because there are too many warning message to stderr in some
        # cases, we must squash them.
        cmd = '/usr/bin/env %s %s 2> /dev/null' % (home_env, cmd)
    else:
        # in debug mode, all messages are redirected to stdout.
        # so we don't have to worry about it.
        cmd = '/usr/bin/env %s %s' % (home_env, cmd)
    [fin, fout, ferr] = os.popen3(cmd)

    """ sweep stderr messages. """
    ####################################
    if (b_debug != 0 or b_cmdline != 0) and os.fork() == 0: # thread processing is cool, but take compatibility.
        # viewarch.b_put_header = 1
        fin.close()
        fout.close()
        errmsg = ""
        errl = ferr.readline()
        while errl:
            if b_cmdline == 1:
                errmsg = errmsg + errl
            else:
                errmsg = errmsg + '[' + tmnow() + '] ' + errl + '<br>'
            errl = ferr.readline()
        ferr.close()

        if b_cmdline == 1:
            print "\n", errmsg, "\n"
        else:
            time.sleep(1)     # :-(
            if errmsg != "":
                print '<p style="color:red">\n'
                print errmsg
                print '</p>\n'
                pass

        sys.exit(0)
    ####################################

    fin.close()
    if wdir != '':
        os.chdir(save_dir)

    return fout

def tsystem0(cmd):

    g_homedir = cfg.general.tla_homedir
    if g_homedir != '':
        home_env = 'HOME=%s ' % g_homedir
    else:
        home_env = ''

    tla_abspath = cfg.general.tla_abspath
    if tla_abspath != '':
        if cmd[:4] == 'tla ':
            cmd = tla_abspath + cmd[3:]

    cmd = '/usr/bin/env %s %s' % (home_env, cmd)
    f = os.system(cmd)

    return f

def tsystem(cmd, wdir=''):

    f = topen(cmd, wdir)
    line = f.readline()
    while line:
        print line,
        line = f.readline()
    f.close()

class tlalog:

    def __init__(self, a, cbv, r = ''):

        if r == '':
            revs = get_revisions(a, cbv)
        else:
            revs = [r]

        self.adic = {}

        for i in range(len(revs)):

            f = topen('tla cat-archive-log -A %s %s--%s' % (a, del_null_branch(cbv), revs[i]))

            ''' read header lines to lines[].'''
            lines = []
            line = f.readline()
            while line[:-1] != '':
                lines.append(line[:-1])
                line = f.readline()

            ''' read body lines to body[].'''
            body = []
            line = f.readline()
            while line:
                body.append(line)
                line = f.readline()
            f.close()

            self.adic[ revs[i] ] = { 'body':body }

            patch_log_list_items = [
                'New-files:',
                'Removed-files:',
                'Modified-files:'
            ]

            key = None
            for j in range(len(lines)):
                if lines[j][0] in [' ', '\t', '\n']:
                    if patch_log_list_items.count(key) > 0:
                        items = string.strip(lines[j])
                        values = string.split(items, ' ')
                    else:
                        values = [ lines[j] ]
                    if not self.adic[ revs[i] ].has_key(key):
                        self.adic[ revs[i] ][ key ] = []
                    self.adic[ revs[i] ][ key ] = self.adic[ revs[i] ][ key ] + values
                else:

                    m = re.match(r"^([^:]+:)\s*(.*)$", lines[j])
                    if m:
                        key = m.group(1)
                        if patch_log_list_items.count(key) > 0:
                            items = string.strip(m.group(2))
                            if items != '':
                                values = string.split(items, ' ')
                                self.adic[ revs[i] ][ key ] = values
                            else:
                                self.adic[ revs[i] ][ key ] = []
                        else:
                            self.adic[ revs[i] ] [ key ] = [string.strip(m.group(2))]
                    else:
                        self.adic[ revs[i] ] [ key ] = []

def patch_log_otherinfo(al):

    filtered_patch_log_list_items = [
        'Keywords:',
        'New-files:',
        'New-directories:',
        'Removed-files:',
        'Removed-directories:',
        'Renamed-files:',
        'Renamed-directories:',
        'Modified-files:',
        'Modified-directories:',
        'New-patches:',
        'Removed-patches:',
        'body',                  # internal
        'Standard-date:',
        'Summary:',
        'Date:',
        'Creator:',
        'Revision:',
        'Archive:'
    ]

    lines = []

    for i in al.keys():
        if i in filtered_patch_log_list_items:
            pass
        else:
            lines.append(i)
            for j in range(len(al[i])):
                lines.append('&nbsp;&nbsp;&nbsp;' + al[i][j])

    return lines

def html_quote(s):

    s = string.replace(s, '>', '&gt;')
    s = string.replace(s, '<', '&lt;')

    return s


def my_default_archive():

    f = topen('tla my-default-archive')
    line = f.readline()
    f.close()

    return line[:-1]

def library_find(a, cbv, r):

    f = topen('tla library-find -A %s %s--%s' % (a, cbv, r))
    line = f.readline()
    status = f.close()
    if status == None:   # if exit with 0, f.close() returns None, not 0.
        return line[:-1]
    else:
        return None

def get_archives(b_filter = 0):

    exclude_archives  = cfg.general.exclude_archives
    exclude_locations = cfg.general.exclude_locations
                                                                                
    f = topen('tla archives')
    a = []
                                                                                
    a_name = f.readline()
                                                                                
    while a_name:
                                                                                
        a_name = string.strip(a_name)
        a_loc  = string.strip(f.readline())
                                                                                
        # If not filtering, or else if the archive is not excluded
        # add the archive name to the archive list
        if (b_filter == 0
            or (re.match(exclude_archives, a_name) == None
                and re.match(exclude_locations, a_loc) == None)):
            a.append(a_name)
                                                                                
        a_name = f.readline()
                                                                                
    f.close()
    return a

def get_categories(a):

    f = topen('tla categories -A %s ' % a)
    c = []
    line = f.readline()
    while line:
        c.append(line[:-1])
        line = f.readline()

    f.close()
    return c

def get_list_common(archive, arg, op):

    f = topen('tla %s -A %s %s' % (op, archive, del_null_branch(arg)))
    list = []
    line = f.readline()
    while line:
        list.append(line[:-1])
        line = f.readline()

    f.close()
    return list

def get_branches(a, c):

    return get_list_common(a, c, 'branches')

def get_versions(a, cb):

    return get_list_common(a, cb, 'versions')

RE_space6_nonsps = re.compile(r"^      (\S+)$")

def get_versions_from_category(a, c):

    f = topen('tla abrowse -A %s %s' % (a, c))
    list = []
    line = f.readline()
    while line:
        m = RE_space6_nonsps.match(line)
        if m != None:
            cbv = add_null_branch(m.group(1))
            list.append(cbv)
        line = f.readline()

    f.close()
    return list

def get_versions_from_archive(a):

    f = topen('tla abrowse --versions -A %s' % a)
    list = []
    line = f.readline()
    while line:
        m = RE_space6_nonsps.match(line)
        if m != None:
            cbv = add_null_branch(m.group(1))
            list.append(cbv)
        line = f.readline()

    f.close()
    return list

def get_all_versions():

    # should be rewrite by generator.
    # "yield" primitive makes it easy, but
    # I still want to stay in Pytho 2.2.

    a_cbv = []
    for a in get_archives():
        for v in get_versions_from_archive(a):
            a_cbv.append([a, v])
    return a_cbv

def matched_archive(a_cbv, exclude_list, b_regexp):

    # b_regexp not implemented yet.

    for i in exclude_list:
        [a, cbv] = i

        if a == None:
            if cbv != None:
                if a_cbv[1] == cbv:
                    return 1
        else:
            if cbv == None:
                if a_cbv[0] == a:
                    return 1
            else:
                if a_cbv == i:
                    return 1
    return 0

def get_filtered_versions(exclude_list, b_regexp):

    filtered_a_cbv = []

    for a_cbv in get_all_revisions():
        if matched_archive(a_cbv, exclude_list, b_regexp) == 0:
            filtered_a_cbv.append(a_cbv)

    return filtered_a_cbv

def get_revisions(a, cbv):

    return get_list_common(a, cbv, 'revisions')

def get_patch(a, cbvr, filename, label):

    if filename[:2] == './':
        filename = filename[2:]

    return get_patch0(a, cbvr, 'patches/%s.patch' % filename, label)

def get_patch_new(a, cbvr, filename, label):

    if filename[:2] == './':
        filename = filename[2:]

    return get_patch0(a, cbvr, 'new-files-archive/%s' % filename, label)

def get_patch_rm(a, cbvr, filename, label):

    if filename[:2] == './':
        filename = filename[2:]

    return get_patch0(a, cbvr, 'removed-files-archive/%s' % filename, label)

def get_patch0(a, cbvr, filename, label):

    tempdir = tempfile.mktemp()
    lines = []
    tsystem('tla get-patch -A %s %s %s' % (a, del_null_branch(cbvr), tempdir))

    if filename[:2] == './':
        filename = filename[2:]

    file = '%s/%s' % (tempdir, filename)

    try:
        f = open(file, 'r')
    except IOError:
        os.system('/bin/rm -rf %s' % tempdir)
        return []

    line = f.readline()
    while line:
        lines.append([label, line])
        line = f.readline()
    f.close()

#   XXX: both routines not worked. why not?
#      os.removedirs(tempdir)
#      os.rmdir(tempdir)

    os.system('/bin/rm -rf %s' % tempdir)

    return lines


def get_patch_all(a, cbvr):

    tempdir = tempfile.mktemp()
    lines = []
    tsystem('tla get-patch -A %s %s %s' % (a, del_null_branch(cbvr), tempdir))

    ''' removedfiles '''
    dir = tempdir + '/removed-files-archive'
    for root, dirs, files in compat.walk(dir):
        for i in range(len(files)):
            r_file = root[ len(dir) : ] + '/' + files[i]
            f = topen('diff -u -L %s -L %s %s /dev/null' % ('orig' + r_file, 'mod' + r_file, root + '/' + files[i]))
            l = f.readlines()
            f.close()
            lines.extend(l)

    ''' newfiles '''
    dir = tempdir + '/new-files-archive'
    for root, dirs, files in compat.walk(dir):
        for i in range(len(files)):
            r_file = root[ len(dir) : ] + '/' + files[i]
            f = topen('diff -u -L %s -L %s /dev/null %s' % ('orig' + r_file, 'mod' + r_file, root + '/' + files[i]))
            l = f.readlines()
            f.close()
            lines.extend(l)

    ''' modifiedfiles '''
    dir = tempdir + '/patches'
    for root, dirs, files in compat.walk(dir):
        for i in range(len(files)):
            f = open(root + '/' + files[i])
            l = f.readlines()
            f.close()
            lines.extend(l)

    os.system('/bin/rm -rf %s' % tempdir)

    return lines

def get_archive_file(a, cbvr, filename, s_seq):

    tempdir = tempfile.mkdtemp()
    tempdir2 = tempdir + '/tmp'

    '''
      format is:
          [s_seq, e_seq, line]

          s_seq ... this line's start revision
          e_seq ... this line's end revision (-1 if still opened)
          line ... content of each line
    '''

    lines = []

    tsystem('tla get -s --no-pristine -A %s %s %s' % (a, del_null_branch(cbvr), tempdir2))
    file = '%s/%s' % (tempdir2, filename)
    f = open(file, 'r')
    line = f.readline()
    while line:
        lines.append([s_seq, -1, line])
        line = f.readline()
    f.close()

#   XXX: both routines not worked. why not?
#      os.removedirs(tempdir)
#      os.rmdir(tempdir)

    os.system('/bin/rm -rf %s' % tempdir)

    return lines


def check_valid_revision(a, c, b, v, r):

    exit_status  = tsystem0("tla cat-archive-log %s/%s--%s--%s--%s > /dev/null 2>&1" % (a, c, b, v, r))

    if exit_status == 0:
        return 1
    else:
        return 0

def get_graph(script_name, v1, v2):

    g_homedir = cfg.general.tla_homedir
    f = topen('python %s/lib/tladdar.py -H %s -c 128 %s %s' % (g_homedir, script_name, v1, v2))
    lines = f.readlines()
    f.close()
    return lines

RE_get_abrowse1 = re.compile(r"^\s+(\S+).*")
RE_get_abrowse2 = re.compile(r"^\s*(?P<date>.+?)\s{2,}(?P<creator>.+?)\s*$")

def get_abrowse(archive, cbv, since_acbvr=None):

    '''
    tla abrowse --reverse --desc -A lord@emf.net--2003b tla--devo--1.1
    lord@emf.net--2003b
      tla
        tla--devo
          tla--devo--1.1
    
            patch-125    (simple changeset)
              Mon Jul 28 00:30:33 PDT 2003      Tom Lord <lord@emf.net>
              small-fixes: ls-archives regexp feature (jblack)
    
            patch-124    (simple changeset)
              Sun Jul 27 23:29:55 PDT 2003      Tom Lord <lord@emf.net>
              sync-tree with jblack past doc patches that wound up on a different project
    
                   ...
    
            patch-1    (simple changeset)
              Sat Jun 21 17:25:02 PDT 2003      Tom Lord <lord@emf.net>
              added some requested aliases
    
            base-0    (tag revsision of lord@emf.net--2003b/tla--devo--1.0--patch-54)
              Sat Jun 21 09:58:31 PDT 2003      Tom Lord <lord@emf.net>
              tag of lord@emf.net--2003b/tla--devo--1.0--patch-54
    
                   ...
    
    '''

    lines = []
    b_since = 0
    if since_acbvr:
        [a, c, b, v, r] = split_acbvr(since_acbvr)
        if archive == a and cbv == '%s--%s--%s' % (c, b, v):
            b_since = 1
            tmpf = '/tmp/,,abr.tmp'   # XXX:
            tmpfp = open(tmpf, 'w')
            tmpfp.write('%s--%s--%s--%s\n' % (c, b, v, r))
            tmpfp.close()

    if b_since:
        f = topen('tla abrowse --reverse --desc --since %s -A %s %s' % (tmpf, archive, del_null_branch(cbv)))
    else:
        f = topen('tla abrowse --reverse --desc -A %s %s' % (archive, del_null_branch(cbv)))

    line = f.readline()
    while line:
        if line[:-1] == '':
            break;
        line = f.readline()

    while 1:
        line = f.readline()
        if line == '':
            break
        m = RE_get_abrowse1.match(line)
        if m != None:
            title = archive + '/' + cbv + '--' + m.group(1)

        line = f.readline()
        if line == '':
            break
	m = RE_get_abrowse2.match(line)
	if m != None:
            date = m.group('date')
	    creator = m.group('creator')

        summary = ''
        line = f.readline()
        while line != '' and line[:-1] != '':
            summary = summary + string.strip(line)
            line = f.readline()

        lines.append([title, creator, date, summary])

    f.close()
    if b_since:
        os.remove(tmpf) # XXX:

    return lines


def get_ancestry(archive, cbv):

    '''
    tez@kamihira.com--2003b
      ViewARCH--tez--0.0.4
        patch-18 ... base-0
      ViewARCH--tez--0.0.3
        patch-31 ... base-0
      ViewArch--tez--0.0.2
        patch-14 ... base-0
    tez@kamihira.com--tmp
      ViewARCH--tez--0.0.1
        patch-16 ... base-0
    '''

    lines = []
    f = topen('tla ancestry -A %s %s' % (archive, del_null_branch(cbv)))
    line = f.readline()
    while line:
        line = line[:-1]
        if len(line) > 0 and line[0] != ' ':
            archive = line
        else:
            if len(line) > 2 and line[:2] == '  ' and line[2] != ' ':
                lines.append([archive, line[2:]])
        line = f.readline()

    f.close()
    return lines

def get_ancestry2(archive, cbv, b_deep_survey, since_acbvr):

    '''
    lord@regexps.com--2002/file-utils--devo--1.0--patch-17	???
    lord@emf.net--2003/file-utils--devo--1.0--base-0	lord@emf.net--...
    lord@emf.net--2003/file-utils--devo--1.0--patch-1	lord@emf.net--...

       ...

    lord@emf.net--2004/tla--devo--1.3--patch-31	lord@emf.net--2004/tla--...
    lord@emf.net--2004/tla--devo--1.3--patch-32	lord@emf.net--2004/tla--...
    '''
    [a, c, b, v, r] = split_acbvr(since_acbvr)
    can_optimize = (archive == a and '%s--%s--%s' % (c, b, v) == cbv)

    if b_deep_survey and not can_optimize:

        lines = []
        f = topen('tla ancestry-graph -r -A %s %s' % (archive, del_null_branch(cbv)))
        line = f.readline()
        while line:
            m = re.match(r"^(\S+)\s+(\S+)", line)
            if m and m.group(2) != "???":
                lines.append(m.group(1))
    
            line = f.readline()
    
        f.close()
        return lines

    else:

        lines = []
        f = topen('tla revisions --full -A %s %s' % (archive, del_null_branch(cbv)))
        line = f.readline()
        while line:
            lines.append(line[:-1])
            line = f.readline()
    
        f.close()
        return lines


def get_last_modified_time(archive, cbv):

    '''
    [tez@bluegate lib]$ tla abrowse --reverse -D tez@kamihira.com--2004s/ViewARCH--devo--0.0.8 | head -10
    tez@kamihira.com--2004s
      ViewARCH
        ViewARCH--devo
          ViewARCH--devo--0.0.8
    
            patch-85
              2004-07-07 06:52:59 GMT
    
            patch-84
              2004-07-07 06:25:50 GMT
    '''

    time = "0000-00-00 00:00:00 GMT"

    f = topen('tla abrowse --reverse -D -A %s %s' % (archive, del_null_branch(cbv)))
    line = f.readline()
    while line:
        m = re.match(r"^ {10,10}(.*)$", line)
        if m:
            time = m.group(1)
            break

        line = f.readline()

    f.close()
    return time

def get_abrowse_creators(archive, cbv):

    lines = []
    f = topen('tla abrowse -c -A %s %s' % (archive, del_null_branch(cbv)))
    line = f.readline()
    seq_no = 0
    while line:
        m = re.match(r"^ {10,10}[^<]*<([^>]*)>.*$", line)
        if m:
            creator = string.strip(m.group(1))
            if len(lines) > 0 and lines[-1][2] == creator:
                lines[-1][1] = seq_no
            else:
                lines.append([seq_no, seq_no, creator])

            seq_no = seq_no + 1

        line = f.readline()

    f.close()
    return lines

def pika_escape(s):

    return pikaesc.pika_esc(s)

def pika_unescape(s):

    return pikaesc.pika_unesc(s)

def _main():

    s = pika_escape("hello world")
    print s
    s = pika_unescape("hello\\(sp)world")
    print s

if __name__ == '__main__': _main()

# arch-tag: 5a563a10-543f-445e-82a1-5b8a609b3e44
