#!/usr/bin/python


################################################################
#
# Copyright (C) 2003,2004 Tez Kamihira
#
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
#

LIB_DIR = None

import sys

if LIB_DIR:
  sys.path.insert(0, LIB_DIR)

import os
import gzt
import re
import urllib
import pwd
import grp
import vcache
import callback
import request
import traceback

try:
	import tempfile
	tempfile.mkstemp
except (ImportError, AttributeError):
	import vtempfile as tempfile

import tlatools
from tlatools import cfg
from tlatools import html_quote
import tladdar
import accept

import summary

import gettext

_ = gettext.gettext


RQ_LIST_ARCHIVE = 0        # just a cgi. no path argument exists.
RQ_LIST_CATEGORY = 1       # category list
RQ_LIST_BRANCH = 2         # branch/version list from category.
RQ_LIST_VERSION = 3        # version list from branch.
RQ_LIST_PATCH_LOG = 4      # show patch log list.
RQ_CAT_PATCH_LOG = 5       # patch log(detail)
RQ_CAT_CSET_FILE = 6       # cs_new, cs_mod, cs_rm
RQ_DRAW_GRAPH = 7          # graph
RQ_ITEM_BROWSE = 8         # tree/file browse.
RQ_GET_ALL_PATCHES = 9     # get all patches
RQ_LIST_INV_HISTORY = 10   # list inventory history
RQ_LIST_ABROWSE = 11       # archive abrowsing instead of RQ_LIST_CATEGORY

VC_VANILLA = 0
VC_BLAME_CHART = 1
VC_CHRONO_CHART = 2
VC_TEXT = 3

'''
[tez@bluegate lib]$ enscript --help-highlight | grep ^Name | cut -c7-
'''

enscript_extensions = {
  '.C' : 'cpp',
  '.EPS' : 'postscript',
  '.DEF' : 'modula_2',  # requires a patch for enscript 1.6.2, see INSTALL
  '.F' : 'fortran',
  '.H' : 'cpp',
  '.MOD' : 'modula_2',  # requires a patch for enscript 1.6.2, see INSTALL
  '.PS' : 'postscript',
  '.S' : 'asm',
  '.SH' : 'sh',
  '.ada' : 'ada',
  '.adb' : 'ada',
  '.ads' : 'ada',
  '.awk' : 'awk',
  '.c' : 'c',
  '.c++' : 'cpp',
  '.cc' : 'cpp',
  '.cpp' : 'cpp',
  '.csh' : 'csh',
  '.cxx' : 'cpp',
  '.diff' : 'diffu',
  '.dpr' : 'delphi',
  '.el' : 'elisp',
  '.eps' : 'postscript',
  '.f' : 'fortran',
  '.for': 'fortran',
  '.gs' : 'haskell',
  '.h' : 'c',
  '.hpp' : 'cpp',
  '.hs' : 'haskell',
  '.htm' : 'html',
  '.html' : 'html',
  '.idl' : 'idl',
  '.java' : 'java',
  '.js' : 'javascript',
  '.lgs' : 'haskell',
  '.lhs' : 'haskell',
  '.m' : 'objc',
  '.m4' : 'm4',
  '.man' : 'nroff',
  '.nr' : 'nroff',
  '.p' : 'pascal',
  # classic setting:
  # '.pas' : 'pascal',
  # most people using pascal today are using the Delphi system originally
  # brought to us as Turbo-Pascal during the eighties of the last century:
  '.pas' : 'delphi',
  # ---
  '.patch' : 'diffu',
  # For Oracle sql packages.  The '.pkg' extension might be used for other
  # file types, adjust here if necessary.
  '.pkg' : 'sql',
  '.pl' : 'perl',
  '.pm' : 'perl',
  '.pp' : 'pascal',
  '.ps' : 'postscript',
  '.py' : 'python',
  '.s' : 'asm',
  '.scheme' : 'scheme',
  '.scm' : 'scheme',
  '.scr' : 'synopsys',
  '.sh' : 'sh',
  '.shtml' : 'html',
  '.sql' : 'sql',
  '.st' : 'states',
  '.syn' : 'synopsys',
  '.synth' : 'synopsys',
  '.tcl' : 'tcl',
  '.tex' : 'tex',
  '.texi' : 'tex',
  '.texinfo' : 'tex',
  '.v' : 'verilog',
  '.vba' : 'vba',
  '.vh' : 'verilog',
  '.vhd' : 'vhdl',
  '.vhdl' : 'vhdl',
}
enscript_filenames = {
  '.emacs' : 'elisp',
  'GNUmakefile' : 'makefile',
  'Makefile' : 'makefile',
  'makefile' : 'makefile',
  'ChangeLog' : 'changelog',
}


def make_enscript_text(lines, f_name):

    # try to figure out the language by examining the filename
    lang_name = ''
    enscript_map = enscript_extensions
    enscript_map.update(enscript_filenames)
    for k in enscript_map.keys():
        if f_name[ - len(k) : ] == k:
            lang_name = enscript_map[k]
            break

    # did the user request to disable this language?
    if lang_name in cfg.general.disable_enscript_lang:
        lang_name = ''

    if lang_name == '':
        return map(html_quote, lines)
    else:

        # i tried this with popen2, but it seemed that some large files got
        # truncated in the buffers somewhere, so i used tempfiles instead to
        # avoid the issue

        (en_out_fd, en_out_name) = tempfile.mkstemp()
        (en_in_fd,  en_in_name)  = tempfile.mkstemp()

        os.write(en_in_fd, "".join(lines))

        os.close(en_out_fd)
        os.close(en_in_fd)

        enscript_exe = cfg.general.enscript_path + 'enscript'
        os.system("%s --color --language=html -E%s -o - %s \
                   | sed -n '/^<PRE>$/,/<\\/PRE>$/p' \
                   > %s"
                   % (enscript_exe, lang_name, en_in_name, en_out_name))

        en_lines = open(en_out_name).readlines()

        # clean up our tempfiles
        os.unlink(en_out_name)
        os.unlink(en_in_name)

        # split the read text on newlines, but preserve all but the last
        # newline, whose presence would cause an annoying trailing blank link in
        # the html pre block
        en_lines[-1] = en_lines[-1][:-1]

        return en_lines


class _item:

    pass


def check_minimum_environment():

    tla_homedir = cfg.general.tla_homedir
    arch_params_dir = tla_homedir + '/.arch-params'
    location_dir = arch_params_dir + '/=locations'

    uid = os.getuid()
    gid = os.getgid()
    uname = "-"
    gname = "-"

    try:
        uname = pwd.getpwuid(uid)[0]
        gname = grp.getgrgid(gid)[0]
    except Exception:
        pass

    # uinfo = '[%s/%s (%d/%d)]' % (uname, gname, uid, gid)
    uinfo = '[%s/%s]' % (uname, gname)

    if not os.access(tla_homedir, os.F_OK or os.R_OK or os.X_OK):
        raise Exception, 'not enough access permission in dir "%s (=tla_homedir)" to %s.' % (tla_homedir, uinfo)
    if not os.access(arch_params_dir, os.F_OK or os.R_OK or os.X_OK):
        raise Exception, 'not enough access permission in dir "%s" to %s.' % (arch_params_dir, uinfo)
    if not os.access(location_dir, os.F_OK or os.R_OK or os.X_OK):
        raise Exception, 'not enough access permission in dir "%s" to %s.' % (location_dir, uinfo)
    for i in os.listdir(location_dir):
        if not os.access(location_dir + '/' + i, os.F_OK or os.R_OK):
            raise Exception, 'not enough access permission in file "%s/%s" to %s.' % (location_dir, i, uinfo)

def chrono_str(s, e):

    if e == None:
        return '%s' % s
    else:
        return '(%s..%s)' % (s, e)

def make_conf_list(conf_str):

    dic = {}
    conf_str = conf_str.strip()
    tmp_list = conf_str.split(',')

    for i in range(len(tmp_list)):
        kv = tmp_list[i].strip()
        m = re.match(r"^(\S+)\s+:\s+('?)(.+)\2$", kv)
        if m:
            k = m.group(1)
            v = m.group(3)
            dic[k] = v

    return dic

def min_max(n, n_min, n_max):

    return min(max(n_min, n), n_max)

def nbsp(n, digits):

    n_fill = digits - len('%d' % n)
    o_str = ''

    if n_fill > 0:
        for i in range(n_fill):
            o_str = o_str + '&nbsp;'

    return o_str

def view_aaaarch(req, total, item_per_page, n_aaaa, pos):

    pg = "pg"

    #url = "http://" + urllib.quote(os.getenv('SERVER_NAME') + os.getenv('SCRIPT_NAME') + os.getenv('PATH_INFO'))
    url = "http://" + req.server_name + req.script_name + req.path_info
    n_page = ( total - 1 ) / item_per_page + 1

    pos = min_max(pos, 1, n_page)
    sp_start = min_max(pos - n_aaaa, 1, n_page)
    sp_end = min_max(pos + n_aaaa - 1, 1, n_page)

    digits = len('%d' % sp_end)

    if sp_start > 1:
        o = '''<table border="0"><tr><td><a href="%s?%s=%d"><b>View</b></a></td>''' % (url, pg, pos - 1)
    else:
        o = '''<table border="0"><tr><td><b>View</b></td>'''

    for i in range(sp_start, sp_end + 1):
        if i == pos:
            o = o + '<td><b>A</b></td>'
        else:
            o = o + '<td><a href="%s?%s=%d">A</a></td>' % (url, pg, i)

    if sp_end < n_page:
        o = o + '''<td><a href="%s?%s=%d"><b>RCH</b></a></td></tr><tr>''' % (url, pg, pos + 1)
    else:
        o = o + '''<td><b>RCH</b></td></tr><tr>'''

    if sp_start > 1:
        o = o + '<td><a href="%s?%s=%d">&lt;&lt;</a></td>' % (url, pg, pos - 1)
    else:
        o = o + '<td></td>'

    for i in range(sp_start, sp_end + 1):
        if i == pos:
            o = o + '<td><b>%d</b></td>' % i
        else:
            o = o + '<td><a href="%s?%s=%d">%d</a></td>' % (url, pg, i, i)

    if sp_end < n_page:
        o = o + '<td align="right"><a href="%s?%s=%d">&gt;&gt;</a></td>' % (url, pg, pos + 1)
    else:
        o = o + '<td></td>'

    o = o + '''</tr></table>'''

    return o


class ls_cat:


    '''
       f ... directory/file name.

    '''

    def __init__(self, a, cbv, r, f):

        self.a = a
        self.cbv = cbv
        self.r = r
        self.f = f

        self.vcache_root = None
        self.revlib_root = None

        ''' setup vcache environment. '''
        vcache_root = cfg.general.tla_cachedir + '/' + a + '/' + cbv
        self.seq = vcache.get_revision_no_from_acbvr(vcache_root, a + '/' + cbv + '--' + r)
        if vcache.rev_available(vcache_root, self.seq):
            self.vcache_root = vcache_root
        else:
            self.vcache_root = None

        ''' setup revlib environment. '''
        self.revlib_root = tlatools.library_find(a, cbv, r)

        self.f_list = None
        self.d_list = None
        self.i_name = None

    def ls(self, b_full):

        if self.vcache_root != None:        
            return self.ls_vcache(b_full)
        if self.revlib_root != None:
            return self.ls_revlib(b_full)

        raise Exception, _("cache not available yet.")

    def ftype(self):

        if self.vcache_root != None:        
            return self.ftype_vcache()
        if self.revlib_root != None:        
            return self.ftype_revlib()

        raise Exception, _("cache not available yet.")

    def get_content(self, v_cmd):

        if self.vcache_root != None and self.i_name != None:        
            return self.get_content_vcache(v_cmd)
        if self.revlib_root != None:
            return self.get_content_revlib(v_cmd)

        raise Exception, _("cache not available yet.")

    def ls_vcache(self, b_full):

        if self.f_list == None:
            chr_f = vcache.load_index(self.vcache_root, 'F')
            self.f_list = chr_f.check_out_to_list(self.seq)

        if self.d_list == None:
            chr_d = vcache.load_index(self.vcache_root, 'D')
            self.d_list = chr_d.check_out_to_list(self.seq)

        f = tlatools.get_files_by_directory_name(self.f_list, self.f, b_full, 'F')
        d = tlatools.get_files_by_directory_name(self.d_list, self.f, b_full, 'D')

        return [f, d]

    def ftype_vcache(self):

        '''
            return [file_type, inventory_id], if possible.
            file_type is either 'F' or 'D'.
            if inventory_id is not available, set to '???'
        '''
        dirp = self.f
        if len(dirp) >= 2 and dirp[:2] == './':
            dirp = dirp[2:]

        if self.f_list == None:
            chr_f = vcache.load_index(self.vcache_root, 'F')
            self.f_list = chr_f.check_out_to_list(self.seq)

        self.i_name = tlatools.find_assoc(self.f_list, dirp)
        if self.i_name == None:
            return 'D'
        else:
            return 'F'

    def get_content_vcache(self, v_cmd):

        return vcache.get_inv_content(self.vcache_root, self.a + '/' + self.cbv + '--' + self.r, self.i_name, v_cmd)

    def ls_revlib(self, b_full):

        # XXX: b_full is not concerned yet.

        dir1 = self.revlib_root
        if len(dir1) >= 1 and dir1[-1] != '/':
            dir1 = dir1 + '/'
        if len(self.f) >= 2 and self.f[:2] == './':
            dir1 = dir1 + self.f[2:]
        else:
            dir1 = dir1 + self.f

        if len(dir1) >= 1 and dir1[-1] != '/':
            dir1 = dir1 + '/'

        files = os.listdir(dir1)
        files.sort()

        f = []
        d = []
        for i in range(len(files)):

            if len(files[i]) >= 2 and files[i][:2] == ',,':
                continue

            if tlatools.f_type(dir1 + files[i]) == 'D':
                d.append(['D', files[i]])
            else:
                f.append(['F', files[i]])

        return [f, d]

    def ftype_revlib(self):

        '''
            return [file_type, inventory_id], if possible.
            file_type is either 'F' or 'D'.
            if inventory_id is not available, set to '???'
        '''
        dir1 = self.revlib_root
        if len(dir1) >= 1 and dir1[-1] != '/':
            dir1 = dir1 + '/'
        if len(self.f) >= 2 and self.f[:2] == './':
            dir1 = dir1 + self.f[2:]
        else:
            dir1 = dir1 + self.f

        if tlatools.f_type(dir1) == 'D':
            return 'D'
        else:
            return 'F'

    def get_content_revlib(self, v_cmd):

        if v_cmd == VC_VANILLA:
            f = self.revlib_root
            if len(f) >= 1 and f[-1] != '/':
                f = f + '/'
            if len(self.f) >= 2 and self.f[:2] == './':
                f = f + self.f[2:]
            else:
                f = f + self.f

            fn = open(f)
            lines = fn.readlines()
            fn.close()
            return lines
        else:
            return ['???', '???', '???']

def rq_list_archive(req, data):

    rows = data['rows']
    rows3 = data['rows3']

    locations = make_conf_list(cfg.general.archive_locations)

    archives = tlatools.get_archives(1)
    for i in range(len(archives)):
        a = _item()
        a.url = req.script_name + '/' + archives[i]
        a.url2 = req.script_name + '/' + archives[i] + '?abr'
        a.title = archives[i]
        a.title2 = '&nbsp;&nbsp;&nbsp;' + locations.get(archives[i], 'null')
        rows.append(a)
    data['title'] = _('Registered archive list')
    data['html_title'] = _('Registered archive list')

    data['url_back'] = ''
    data['url_back_title'] = ''

    sm = summary.summary(cfg.general.tla_cachedir + '/,,summary.txt')
    sm_len = int(cfg.general.summary_length)
    tlist = sm.getlist()
    sm_len = min(len(tlist), sm_len)
    for i in range(sm_len):
        a = _item()
        a.time = tlist[i][0]
        a.acbv = tlist[i][1]
        a.acbv_url = req.script_name + '/' + a.acbv
        vcache_dir = cfg.general.tla_cachedir + '/' + a.acbv
        b_dsm = vcache.is_dsm(vcache_dir)
        if b_dsm:
            a.acbv_url = req.script_name + '/' + a.acbv + '--dsm'
        rows3.append(a)

    return cfg.general.list

def rq_list_abrowse(req, data):

    n_categories = 0
    n_branches = 0
    n_versions = 0

    lines = []
    f = tlatools.topen('tla abrowse -A %s' % req.archive)
    l = f.readline()
    while l:
        m = re.match(r"^(\s*)(.+)$", l)
        if m:
            spaces = m.group(1)
            nbsps = '&nbsp;' * len(spaces) * 3
            item = m.group(2)
            if len(spaces) == 0:    # archive line
                l_link = '<a href="%s/%s">%s</a>' % (req.script_name, item, item)
                l_append = nbsps + l_link + '\n'
                c_a = item
            elif len(spaces) == 2:  # category line
                l_link = '<a href="%s/%s/%s">%s</a>' % (req.script_name, c_a, item, item)
                l_append = nbsps + l_link + '\n'
                c_c = item
                n_categories = n_categories + 1
            elif len(spaces) == 4:  # branch line
                item2 = item.split('--')[1]
                l_link = '<a href="%s/%s/%s">%s</a>' % (req.script_name, c_a, item, item2)
                l_append = nbsps + l_link + '\n'
                c_cb = item
                n_branches = n_branches + 1
            elif len(spaces) == 6:  # version line

                vcache_dir = cfg.general.tla_cachedir + '/' + req.archive + '/' + item
                b_dsm = vcache.is_dsm(vcache_dir)

                item2 = item.split('--')[2]
                l_link = '<a href="%s/%s/%s">%s</a>' % (req.script_name, c_a, item, item2)
                if b_dsm:
                    l_link2 = '<a href="%s/%s/%s--dsm">(dsm)</a>' % (req.script_name, c_a, item)
                    l_append = nbsps + l_link + '&nbsp;&nbsp;' + l_link2 + '\n'
                else:
                    l_append = nbsps + l_link + '\n'
                c_cbv = item
                n_versions = n_versions + 1
            elif len(spaces) == 8:  # revision line
                revs = item.split(' .. ') # lazy...
                if len(revs) == 1:
                    l_link = '<a href="%s/%s/%s--%s">%s</a>' % (req.script_name, c_a, c_cbv, revs[0], revs[0])
                    l_append = nbsps + l_link + '\n'
                elif len(revs) == 2:
                    l_link = '<a href="%s/%s/%s--%s">%s</a>' % (req.script_name, c_a, c_cbv, revs[0], revs[0])
                    l_link2 = '<a href="%s/%s/%s--%s">%s</a>' % (req.script_name, c_a, c_cbv, revs[1], revs[1])
                    l_append = nbsps + l_link + ' .. ' + l_link2 + '\n'
                else:
                    l_append = l
            else:
                l_append = l
        else:
            l_append = l
        lines.append(l_append)
        l = f.readline() 
    f.close()

    data['n_categories'] = str(n_categories)
    data['n_branches'] = str(n_branches)
    data['n_versions'] = str(n_versions)
    data['lines'] = lines
    data['html_title'] = 'abrowse list of %s' % req.archive
    data['title'] = data['html_title']

    data['back_url'] = ''
    data['cbvr'] = ''

    return cfg.general.abrowse

def rq_list_category(req, data):

    rows = data['rows']

    categories = tlatools.get_categories(req.archive)
    for i in range(len(categories)):
        a = _item()
        a.title = categories[i]
        a.title2 = ''
        a.url = req.script_name + '/' + req.archive + '/' + a.title
        rows.append(a)
    data['title'] = _('Categories for archive <strong>%s</strong>') % req.archive
    data['html_title'] = req.archive

    data['url_back'] = req.script_name
    data['url_back_title'] = _('Back to archives')

    return cfg.general.list

def rq_list_branch(req, data):

    rows = data['rows']

    descriptions = make_conf_list(cfg.general.version_descriptions)

    if cfg.general.tla_skip_branch_browsing:

        versions = tlatools.get_versions_from_category(req.archive, req.category)
        title_str = _('Versions')

        for i in range(len(versions)):
            vcache_dir = cfg.general.tla_cachedir + '/' + req.archive + '/' + versions[i]
            b_dsm = vcache.is_dsm(vcache_dir)
            a = _item()
            a.title = versions[i]
            if b_dsm:
                a.title = a.title + '*' 
            a.title2 = descriptions.get(versions[i], '')
            a.url = req.script_name + '/' + req.archive + '/' + versions[i]
            if b_dsm:
                a.url = a.url + '--dsm'
            rows.append(a)

    else:
        branches = tlatools.get_branches(req.archive, req.category)
        title_str = _('Branches')

        for i in range(len(branches)):
            a = _item()
            a.title = branches[i]
            a.title2 = descriptions.get(branches[i], '')
            a.url = req.script_name + '/' + req.archive + '/' + a.title
            rows.append(a)

    data['title'] = _('%s for category <strong>%s/%s</strong>') % (title_str, req.archive, req.category)
    data['html_title'] = req.archive + '/' + req.category
    data['url_back'] = req.script_name + '/' + req.archive
    data['url_back_title'] = _('Back to categories')

    return cfg.general.list

def rq_list_version(req, data):

    rows = data['rows']

    versions = tlatools.get_versions(req.archive, req.category + '--' + req.branch)
    for i in range(len(versions)):
        a = _item()
        a.title = versions[i]
        a.title2 = ''
        a.url = req.script_name + '/' + req.archive + '/' + a.title
        rows.append(a)
    data['title'] = _('Versions for branch <strong>%s/%s--%s</strong>') % (req.archive, req.category, req.branch)
    data['html_title'] = req.archive + '/' + req.category + '--' + req.branch

    data['url_back'] = req.script_name + '/' + req.archive + '/' + req.category
    data['url_back_title'] = _('Back to branches')

    return cfg.general.list

def rq_list_patch_log(req, data):

    rows = data['rows']
    rows2 = data['rows2']
    rows3 = data['rows3']

    opt_paging = cfg.general.paging
    b_paging = 0
    if opt_paging != '':
        b_paging = 1
        [opt_pg1, opt_pg2] = opt_paging.split(',')
        opt_pg1 = int(opt_pg1)
        opt_pg2 = int(opt_pg2)
        pg_cur = int(req.params.get('pg', '1'))
        pg_start = opt_pg1 * (pg_cur - 1)

    version = '--'.join([req.category, req.branch, req.version])
    vcache_dir = cfg.general.tla_cachedir + '/' + req.archive + '/' + version

    if req.b_dsm == 1:
        abrowse_list = vcache.get_abrowse(vcache_dir)
        [max_rev_plus_1, dummy] = vcache.get_state(vcache_dir)
    else:
        abrowse_list = tlatools.get_abrowse(req.archive, version)
        max_rev_plus_1 = vcache.get_matched_rev_count(vcache_dir, req.archive + '/' + version)
    abrowse_len = len(abrowse_list)

    for i in range(abrowse_len):
        if b_paging == 0 or (i >= pg_start and i < pg_start + opt_pg1):
            a = _item()
            a.title = callback.revlist_header_cb(abrowse_len - i - 1, abrowse_list[i][0], 0, 0)
            if req.b_dsm == 1:
                a.url = req.script_name + '/' + req.archive + '/' + version + '--dsm-' + str(abrowse_len - i - 1)
            else:
                a.url = req.script_name + '/' + abrowse_list[i][0]
            if (abrowse_len - i) > max_rev_plus_1:
                a.cache_ok = ''
            else:
                a.cache_ok = '1'

            [a.creator, a.email] = tlatools.split_creator_email(abrowse_list[i][1])
            a.date = tlatools.round_sec(abrowse_list[i][2])
            a.summary = html_quote(''.join(abrowse_list[i][3]))
            a.filename = ''

            rows.append(a)

    data['acbv'] = req.archive + '/' + version

    data['url_book_mark'] = req.script_name + '/' + req.archive + '/' + version

    data['rev2'] = ''
    data['url_show_graph'] = ''

    data['url_back_versions'] = req.script_name + '/' + req.archive + '/' + req.category + '--' + req.branch

    data['title'] = _('Revisions for version <strong>%s</strong>') % data['acbv']
    data['html_title'] = data['acbv']

    if b_paging == 1 and len(abrowse_list) > opt_pg1:
        data['view_aaaarch'] = view_aaaarch(req, len(abrowse_list), opt_pg1, opt_pg2, pg_cur)
    else:
        data['view_aaaarch'] = ''

    return cfg.general.revisions

def rq_cat_patch_log(req, data):

    version = '--'.join([req.category, req.branch, req.version])

    if req.b_dsm == 1:
        vcache_dir = cfg.general.tla_cachedir + '/' + req.archive + '/' + version
        acbvr = req.archive + '/' + version + '--' + req.revision
        acbvr_2 = vcache.v2p_acbvr(vcache_dir, acbvr)
        [ra, rc, rb, rv, rr] = tlatools.split_acbvr(acbvr_2)
        al = tlatools.tlalog(ra, '%s--%s--%s' % (rc, rb, rv), rr)
        al = al.adic[ rr ]
    else:
        al = tlatools.tlalog(req.archive, version, req.revision)
        al = al.adic[ req.revision ]

    data['back_url'] = req.script_name + '/' + req.archive + '/' + version
    data['op_url'] = req.script_name + '/' + req.archive + '/' + version + '--' + req.revision
    data['acbvr'] = req.archive + '/' + version + '--' + req.revision
    data['cbvr'] = version + '--' + req.revision
    data['creator'] = html_quote(' '.join(al['Creator:']))
    data['date'] = ' '.join(al['Standard-date:'])
    data['patch_level'] = req.revision

    data['body'] = map(html_quote, al['body'])

    data['revision'] = ' '.join(al['Revision:'])
    data['summary'] = html_quote(' '.join(al['Summary:']))

    data['keywords'] = al.get('Keywords:', [])
    data['newfiles'] = \
      map(lambda x: '<A href="%s?cs_new=%s">%s</A>' % (data['op_url'], urllib.quote(x), x),
                      al.get('New-files:', []))

    data['newdirectories'] = al.get('New-directories:', [])
    data['removedfiles'] = \
      map(lambda x: '<A href="%s?cs_rm=%s">%s</A>' % (data['op_url'], urllib.quote(x), x),
                      al.get('Removed-files:', []))

    data['removeddirectories'] = al.get('Removed-directories:', [])
    data['renamedfiles'] = al.get('Renamed-files:', [])
    data['renameddirectories'] = al.get('Renamed-directories:', [])

    data['modifiedfiles'] = \
      map(lambda x: '<A href="%s?cs_mod=%s">%s</A>' % (data['op_url'], urllib.quote(x), x),
                      al.get('Modified-files:', []))

    data['modifieddirectories'] = al.get('Modified-directories:', [])
    new_patches = al.get('New-patches:', [])
    tladdar.contract_list(new_patches, 1)
    data['newpatches'] = new_patches
    data['removedpatches'] = al.get('Removed-patches:', [])
    data['html_title'] = data['acbvr']
    data['title'] = _('Patch set for revision <strong>%s</strong>') % data['acbvr']

    data['otherinfo'] = tlatools.patch_log_otherinfo(al)

    return cfg.general.patch_set

def rq_cat_cset_file(req, data):

    cbv = '--'.join([req.category, req.branch, req.version])
    cbvr = cbv + '--' + req.revision
    acbv = req.archive + '/' + cbv
    acbvr = req.archive + '/' + cbvr
    data['acbvr'] = acbvr
    data['back_url_1'] = req.script_name + '/' + acbvr
    data['back_url_2'] = req.script_name + '/' + acbv

    # data['cs_file'] = req.params['file']

    if req.b_dsm == 1:
        vcache_dir = cfg.general.tla_cachedir + '/' + acbv
        acbvr_2 = vcache.v2p_acbvr(vcache_dir, acbvr)

        [a, c, b, v, r] = tlatools.split_acbvr(acbvr_2)

        gp_a = a
        gp_cbvr = '%s--%s--%s--%s' % (c, b, v, r)
    else:
        gp_a = req.archive
        gp_cbvr = cbvr

    if req.params.has_key('cs_new'):
        data['cs_file'] = req.params.get('cs_new', '')
        lines = tlatools.get_patch_new(gp_a, gp_cbvr, data['cs_file'], 'dummy')
        data['title'] = _('New file: <strong>%s</strong>') % data['cs_file']
    elif req.params.has_key('cs_mod'):
        data['cs_file'] = req.params.get('cs_mod', '')
        lines = tlatools.get_patch(gp_a, gp_cbvr, data['cs_file'], 'dummy')
        data['title'] = _('Modified file: <strong>%s</strong>') % data['cs_file']
    elif req.params.has_key('cs_rm'):
        data['cs_file'] = req.params.get('cs_rm', '')
        lines = tlatools.get_patch_rm(gp_a, gp_cbvr, data['cs_file'], 'dummy')
        data['title'] = _('Removed file: <strong>%s</strong>') % data['cs_file']
    else:  # ???
        data['cs_file'] = ''
        lines = tlatools.get_patch(gp_a, gp_cbvr, data['cs_file'], 'dummy')

    data['cs_contents'] = map(lambda x: html_quote(x[1]), lines)
    data['html_title'] = acbvr + '(diff)'

    return cfg.general.cs

def rq_draw_graph(req, data):

    rev1 = req.archive + '/' + req.category + '--' + req.branch + '--' + req.version
    rev2 = req.params.get('graph', '')

    '''
    http://localhost/cgi-bin/viewarch/viewarch.cgi/lord@emf.net--2003b/tla--devo--1.1?
           graph=stig@brautaset.org--arwen-03/file-utils--stig--1.0
    '''
    lines = tlatools.get_graph(req.script_name, rev1, rev2)
    data['lines'] = lines
    data['html_title'] = rev1 + ' & ' + rev2
    data['title'] = rev1 + '&nbsp;&nbsp; & &nbsp; &nbsp; ' + rev2

    return cfg.general.plain

def rq_item_browse(req, data):

    rows = data['rows']
    rows2 = data['rows2']
    rows3 = data['rows3']

    cbv = req.category + '--' + req.branch + '--' + req.version
    acbv = req.archive + '/' + cbv

    lscat = ls_cat(req.archive, cbv, req.revision, req.dirpart)

    f_type = lscat.ftype()

    if f_type == 'F':     # cat normal file.

        vv_cmd = { 'bc':VC_BLAME_CHART, 'cc':VC_CHRONO_CHART, 'text':VC_TEXT }
        v_cmd = VC_VANILLA
        for k in vv_cmd.keys():
            if req.params.has_key(k):
                v_cmd = vv_cmd[k]
                break

        if v_cmd == VC_TEXT:  # if text download then set vanilla mode.
            lines = lscat.get_content(VC_VANILLA)
        else:
            lines = lscat.get_content(v_cmd)

        if v_cmd == VC_VANILLA:   # vanilla (including syntax-highlighting)
            if data['opt_use_enscript'] == 1:
                data['lines'] = make_enscript_text(lines, req.dirpart)
            else:
                data['lines'] = map(html_quote, lines)

        elif v_cmd == VC_BLAME_CHART:  # blame
            vcache_dir = cfg.general.tla_cachedir + '/' + acbv
            rmap = vcache.load_revmap(vcache_dir)

            """phase 1: check max fields length."""
            max_len_0 = 0
            for i in range(len(lines)):
    
                line_acbvr = rmap.getitem(lines[i][0])
                x0 = callback.line_header_cb(lines[i][0], line_acbvr, 0, 0)
                max_len_0 = max(len(x0), max_len_0)

            """phase 2: actually draw it."""
            o_line = []
            for i in range(len(lines)):
                line_acbvr = rmap.getitem(lines[i][0])
                x0 = callback.line_header_cb(lines[i][0], line_acbvr, 0, 0)
                x1 = lines[i][1]
                o_line.append( html_quote('%*.*s : %s' % (max_len_0, max_len_0, x0, x1)))

            data['lines'] = o_line

        elif v_cmd == VC_CHRONO_CHART:  # chrono
            vcache_dir = cfg.general.tla_cachedir + '/' + acbv
            rmap = vcache.load_revmap(vcache_dir)

            """phase 1: check max fields length."""
            max_len_0 = 0
            max_len_1 = 0
            for i in range(len(lines)):

                line_acbvr = rmap.getitem(lines[i][0])
                x0 = callback.line_header_cb(lines[i][0], line_acbvr, 0, 0)
                max_len_0 = max(len(x0), max_len_0)

                if lines[i][1] != -1:
                    line_acbvr = rmap.getitem(lines[i][1])
                    x1 = callback.line_header_cb(lines[i][1], line_acbvr, 0, 0)
                    max_len_1 = max(len(x1), max_len_1)

            """phase 2: actually draw it."""
            max_len_total = max_len_0 + max_len_1 + 5    # 5 is XXX:
            o_line = []
            for i in range(len(lines)):
                line_acbvr = rmap.getitem(lines[i][0])
                x0 = callback.line_header_cb(lines[i][0], line_acbvr, 0, 0)
                if lines[i][1] != -1:
                    line_acbvr = rmap.getitem(lines[i][1])
                    x1 = callback.line_header_cb(lines[i][1], line_acbvr, 0, 0)
                else:
                    x1 = None
                x2 = lines[i][2]
                line = html_quote('%*.*s : %s' % (max_len_total, max_len_total, chrono_str(x0, x1), x2))
                if x1:
                    line = '<div style="background-color:gray">' + line[:-1] + '</div>'
                else:
                    line = '<div>' + line[:-1] + '</div>'
                o_line.append(line)

            data['lines'] = o_line

        elif v_cmd == VC_TEXT:  # text download
            put_header(1)
            print ''.join(lines)
            sys.exit(0)

        data['html_title'] = req.archive + '/' + cbv + '/' + req.revision + '/' + req.dirpart[2:]
        data['title'] = req.dirpart

        return cfg.general.plain

    else:    # directory browsing.

        b_full = req.params.has_key('full')
        [f_files, d_files] = lscat.ls(b_full)

        url_head = "%s/%s/%s--%s--%s" % (req.script_name, req.archive, req.category, req.branch, req.version)

        dirp = req.dirpart
        if len(dirp) >= 2 and dirp[:2] == './':
            dirp = dirp[2:]

        if len(dirp) > 0 and dirp[-1] != '/':
            dirp = dirp + '/'

        f_d_files = d_files + f_files

        if b_full:
            f_d_files.sort(lambda x, y: cmp(x[1], y[1]))
        for i in range(len(f_d_files)):
            a = _item()
            a.title = f_d_files[i][1]
            a.isdir = ''
            if f_d_files[i][0] == 'D':  # directory.
                a.title = a.title + '/'
                a.isdir = '1'
            a.title2 = ''
            if b_full:
                a.url = urllib.quote("%s--%s/%s" % (url_head, req.revision, f_d_files[i][1]))
            else:
                a.url = urllib.quote("%s--%s/%s%s" % (url_head, req.revision, dirp, f_d_files[i][1]))
            rows2.append(a)

        data['title'] = _('Directory <strong>%s</strong>') % req.dirpart
        data['html_title'] = _('Directory ') + req.dirpart

        data['url_back'] = ''
        data['url_back_title'] = ''

        return cfg.general.list

def rq_get_all_patches(req, data):

    cbv = req.category + '--' + req.branch + '--' + req.version
    cbvr = cbv + '--' + req.revision

    if req.b_dsm == 1:
        acbv = req.archive + '/' + cbv
        acbvr = req.archive + '/' + cbvr
        vcache_dir = cfg.general.tla_cachedir + '/' + acbv
        acbvr_2 = vcache.v2p_acbvr(vcache_dir, acbvr)

        [a, c, b, v, r] = tlatools.split_acbvr(acbvr_2)

        lines = tlatools.get_patch_all(a, "%s--%s--%s--%s" % (c, b, v, r))
    else:
        lines = tlatools.get_patch_all(req.archive, cbvr)

    lines2 = []
    for i in range(len(lines)):
        line = html_quote(lines[i])
        if line[0] == '-':
            line = '<div style="background-color:darkgray">' + line + '</div>'
        elif line[0] == '+':
            line = '<div style="background-color:lightblue">' + line + '</div>'
        elif line[0] == '\\':
            line = '<div style="background-color:gray">' + line + '</div>'
        else:
            line = '<div>' + line + '</div>'
        lines2.append(line)

    data['lines'] = lines2
    data['html_title'] = cbvr
    data['title'] = '<strong>%s</strong>' % cbvr

    return cfg.general.plain

def rq_list_inv_history(req, data):

    rows = data['rows']

    cbv = req.category + '--' + req.branch + '--' + req.version
    cbvr = cbv + '--' + req.revision

    vcache_dir = cfg.general.tla_cachedir + '/' + req.archive + '/' + cbv
    iv_name = vcache.get_inv_name(vcache_dir, req.archive + '/' + cbvr, req.dirpart)

    abrowse_list = vcache.get_abrowse(vcache_dir)
    cf_f = vcache.load_inv_by_iname(vcache_dir, iv_name)

    meta_info = cf_f.get_meta_info()
    for j in range(len(meta_info)):

        [commit_rev, f_name, f_type] = meta_info[-(j+1)]

        a = _item()
        abr_idx = - (commit_rev + 1)
        rmap = vcache.load_revmap(vcache_dir)
        acbvr = rmap.getitem(commit_rev)
        if req.b_dsm == 1:
            a.title2 = req.archive + '/' + cbv + '--dsm-' + str(commit_rev)
        else:
            a.title2 = acbvr
        a.title = callback.invhist_header_cb(commit_rev, acbvr, 0, 0)
        a.url = urllib.quote(req.script_name + '/' + a.title2)

        [a.creator, a.email] = tlatools.split_creator_email(abrowse_list[j][1])
        a.date = tlatools.round_sec(abrowse_list[abr_idx][2])
        a.summary = html_quote(''.join(abrowse_list[abr_idx][3]))
        a.filename = f_name

        a.cache_ok = '1'

        rows.append(a)

    data['acbv'] = req.archive + '/' + cbv

    data['url_book_mark'] = req.script_name + '/' + req.archive + '/' + cbv

    data['rev2'] = ''
    data['url_show_graph'] = ''

    data['url_back_versions'] = req.script_name + '/' + req.archive + '/' + req.category + '--' + req.branch

    data['title'] = _('Inventory: &nbsp;&nbsp;&nbsp; <strong>%s</strong>') % iv_name
    data['html_title'] = _('Inventory: %s') % iv_name
    data['view_aaaarch'] = ''

    return cfg.general.revisions

def _main():

    req = request.request()

    #os.environ['LANG'] = 'ja_JP.eucJP'
    os.environ['LANG'] = req.language

    gettext.bindtextdomain('messages', cfg.general.tla_homedir + '/locale')
    gettext.textdomain("messages")

    check_minimum_environment()

    if req.script_name == None:
        req.script_name = '??'

    data = {
        'req':               req,
        'cfg':               cfg,
        'rows':              [],
        'rows2':             [],
        'rows3':             [],
        'contents':          [],
        'vsn':               '0.0.12',
        'navigator':         req.make_navigator(),
        'script_name':       req.script_name,
        'opt_second_pick':   None,
        'opt_use_enscript':  None
    }

    if cfg.general.second_pick == '1':
        data['opt_second_pick'] = 1

    if cfg.general.use_enscript == '1':
        data['opt_use_enscript'] = 1

    data['req_type'] = req.type

    template_name = {
      RQ_LIST_ARCHIVE:      rq_list_archive,
      RQ_LIST_CATEGORY:     rq_list_category,
      RQ_LIST_BRANCH:       rq_list_branch,
      RQ_LIST_VERSION:      rq_list_version,
      RQ_LIST_PATCH_LOG:    rq_list_patch_log,
      RQ_CAT_PATCH_LOG:     rq_cat_patch_log,
      RQ_CAT_CSET_FILE:     rq_cat_cset_file,
      RQ_DRAW_GRAPH:        rq_draw_graph,
      RQ_ITEM_BROWSE:       rq_item_browse,
      RQ_GET_ALL_PATCHES:   rq_get_all_patches,
      RQ_LIST_INV_HISTORY:  rq_list_inv_history,
      RQ_LIST_ABROWSE:      rq_list_abrowse,
    }[req.type](req, data)

    put_header()

    try:
        tn = template_name.replace('%lang%', req.language)
        template = gzt.Template(os.path.join(tlatools.pathinfo_install_dir(), tn))
    except Exception:
        tn = template_name.replace('%lang%', 'default')
        template = gzt.Template(os.path.join(tlatools.pathinfo_install_dir(), tn))

    template.generate(sys.stdout, data)

b_put_header = 0

def put_header(b_plain = 0):

    global b_put_header

    if b_put_header == 0:
        if b_plain == 1:
            print 'Content-type: text/plain\n\n'
        else:
            print 'Content-type: text/html\n\n'
        b_put_header = 1

def _main0():

    try:
        _main()

    except SystemExit, inst:

        if str(inst) != "0":
            put_header()
            print '''<html><head></head><body><pre>'''
            traceback.print_exc(file=sys.stdout)
            print '''</pre></body></html>'''

    except Exception, inst:

        put_header()
        print '''<html><head></head><body><pre>'''
        traceback.print_exc(file=sys.stdout)
        print '''</pre></body></html>'''

if __name__ == '__main__': _main0()

# arch-tag: 34c8f768-7a0e-48d3-8308-722d5564afba
