#!/usr/bin/python

################################################################
#
# Copyright (C) 2004 Tez Kamihira
#
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
#

################################################################
#
# vsummary.py -- version summary.
#

import re
import sys
import tlatools
import os
import stat
import string

class vsummary:

    def __init__(self, in_file):

        self.load(in_file)
        self.append_bf = ""

    def load(self, in_file):

        self.in_file = in_file

    def getsize(self):

        f = open(self.in_file)
        size = 0
        l = f.readline()
        while l != "":
            m = re.match(r"^(\d+)", l)
            count = int(m.group(1))
            for i in range(count):
                f.readline()
            size = size + 1
            l = f.readline()

        f.close()

        return size

    def get_patch(self, n):

        f = open(self.in_file)
        curr_n = 0
        l = f.readline()
        while l != "" and curr_n < n:
            m = re.match(r"^(\d+)", l)
            count = int(m.group(1))
            for i in range(count):
                f.readline()
            curr_n = curr_n + 1
            l = f.readline()

        m = re.match(r"^(\d+)", l)
        count = int(m.group(1))
        out = []
        for i in range(count):
            out.append(f.readline())
        f.close()

        return out

    def get_selective_fields(self, fields, min, max):

        '''
        result is an array each of which element is, again, array.
        the size of the array is (max - min + 1).
        '''

        ''' open vsummary file. '''
        f = open(self.in_file)

        ''' skip previous lines. '''
        curr_n = 0
        l = f.readline()
        while l != "" and curr_n < min:
            m = re.match(r"^(\d+)", l)
            count = int(m.group(1))
            for i in range(count):
                f.readline()
            curr_n = curr_n + 1
            l = f.readline()

        ''' main part. '''
        out = []
        while l != "" and curr_n < max + 1:
            m = re.match(r"^(\d+)", l)
            count = int(m.group(1))

            ''' read next patch log into list. '''
            patch_lines = []
            b_body = 0
            body = ""
            for i in range(count):
                l = f.readline()
                if b_body == 0:
                    if l[0] == '\n':
                        b_body = 1
                    else:
                        patch_lines.append(l)
                else:
                    body = body + l

            ''' analize the contents. '''
            dic = {}
            dic['=body'] = body
            i = 0
            while i < len(patch_lines):
                m = re.match(r"^([^ \t:]+:)(.*)$", patch_lines[i])
                if m:
                    key = m.group(1)
                    val = m.group(2)
                    dic[key] = val
                else:
                    dic[key] = dic[key] + patch_lines[i]
                i = i + 1

            ''' extract the result. '''
            field_list = []
            for i in range(len(fields)):
                field_list.append(dic.get(fields[i], '<<<no summary: header>>>'))
            out.append(field_list)

            curr_n = curr_n + 1
            l = f.readline()

        ''' epilogue. '''
        f.close()

        return out

    def save(self, out_file, append_data):

        bufsz = 64 * 1024

        g = open(out_file, 'w')

        try:
            f = open(self.in_file)
            s_in = f.read(bufsz)
            while (s_in != ""):
                g.write(s_in)
                s_in = f.read(bufsz)
            f.close()
        except IOError:
            pass

        if append_data != "":
            if append_data[-1] != '\n':
                append_data = append_data + '\n'

            n = append_data.count('\n')
            if n == 1:
                plural = ""
            else:
                plural = "S"
            g.write("%d LINE%s #################################################\n" % (n, plural))
            g.write(append_data)

        g.close()

def _main():

    # def get_selective_fields(self, fields, min, max):
    '''
    Revision: ViewARCH--devo--0.0.9--patch-60
    Archive: tez@kamihira.com--2004s
    Creator: Tez Kamihira <tez@kamihira.com>
    Date: Fri Dec  3 15:25:23 JST 2004
    Standard-date: 2004-12-03 06:25:23 GMT
    Modified-files: lib/gauge_rev.py lib/gencache.py
        lib/vcache.py
    New-patches: tez@kamihira.com--2004s/ViewARCH--devo--0.0.9--patch-60
    Summary: breath.
    Keywords: 

    '''

    v = vsummary("./tmp/tez@kamihira.com--2004s/ViewARCH--devo--0.0.9/control/vsummary.log")
    n = v.getsize()
    print n
    o = v.get_selective_fields(['Revision:', 'Archive:', 'Creator:'], 190, 202)
    print o

if __name__ == '__main__': _main()

# arch-tag: a495daea-1e51-4f60-b4c7-6cbe2b0a2ab7
